From iris.bi Require Import bi.
From iris.bi Require Export fractional.
From iris.algebra Require Import auth numbers.
From iris.proofmode Require Import tactics.
From iris.base_logic Require Import base_logic own.

From diaframe Require Import proofmode_base.
From diaframe.lib Require Import tokenizable own_hints.


(* This file defines the token ghost state used by, among others, ARC.
   The concrete definitions are abstracted away in this library too provide
   a cleaner presentation of the required ghost-state reasoning.
   Additionally, the proofs of the various updates sometimes require foresight, 
   and Diaframe cannot handle such proofs automatically. 
   Since this Diaframe comes with biabduction hints, Diaframe can reason about token automatically. *)

Section fractional_tokenizable.
  Definition fracTokenR := authUR $ optionUR $ prodR fracR positiveR.
  Class fracTokenG Σ := { fracPayloadTokenCounter_inG :> inG Σ fracTokenR }.
  Definition fracTokenΣ : gFunctors := #[GFunctor fracTokenR].
  Global Instance subG_fracTokenΣ {Σ} : subG fracTokenΣ Σ → fracTokenG Σ.
  Proof. solve_inG. Qed.

  Context `{!fracTokenG Σ}.
  Context (P : Qp → iProp Σ).
  Context {HP : Fractional P}.

  Instance P_AsFractional q : AsFractional (P q) P q.
  Proof using HP. done. Qed.

  Program Definition fractional_payload_token_counter : tokenizable.Tokenizable $ iPropI Σ :=
    tokenizable.Build_Tokenizable
      (P 1)
      (λ γ, ∃ q, own γ (◯ (Some (q, 1%positive))) ∗ P q)%I
      (λ γ p, ∃ q q', own γ (● (Some (q, p))) ∗ ⌜(q + q' = 1)%Qp⌝ ∗ P q')%I
      (λ γ q, own γ (●{#q} None))
      _.
  Next Obligation.
    split.
    - iStepsS.
    - do 3 iStepS.
      iStepDebug.
      solveStep.
      solveStep.
      iExists (x1 + (x2 / 2))%Qp.
      unshelve solveSteps; [shelve | | split ].
      { apply frac_valid. rewrite -H. apply Qp.add_le_mono_l. by apply Qp.lt_le_incl, Qp.div_lt. }
      { rewrite -H -assoc.
        apply Qp_add_eq_mono_l, Qp.div_2. }
      iDestruct "H2" as "[H2 H2']".
      replace (Pos.succ x0 - x0)%positive with 1%positive by lia.
      iStepsS.
    - do 2 iStepS.
      iStepDebug.
      solveStep.
      solveStep.
      iExists (1/2)%Qp.
      solveSteps.
      split. { apply Qp.div_2. }
      iDestruct "H2" as "[H2 H2']".
      iStepsS.
    - do 4 iStepS.
      apply Qp.lt_sum in H2 as [q ->].
      iExists q.
      iStepS.
      iExists (x4 + x3)%Qp.
      iStepsS.
      * by rewrite assoc (Qp.add_comm q).
      * by iCombine "H4 H2" as "$".
    - iStepsS.
      rewrite -H.
      by iCombine "H4 H2" as "$".
    - do 3 iStepS.
    - do 4 iStepS.
    - iSolveTC.
    - do 3 iStepS. iDestruct (own_valid with "H1") as %[Hx0%dfrac_valid_own _]%auth_auth_dfrac_valid.
      revert Hx0 => /dfrac_valid_own Hx0; iStepsS.
    - intros γ p q. eapply (anti_symm _); iStepS.
      * by iDestruct "H1" as "[$ $]".
      * by rewrite Qp.add_comm.
  Qed.

  Let tc := fractional_payload_token_counter.

  Definition token := tokenizable.token tc.
  Definition token_counter := tokenizable.token_counter tc.
  Definition no_tokens := tokenizable.no_tokens tc.
End fractional_tokenizable.

Section automation.
  Context `{!fracTokenG Σ}.
  Context (P : Qp → iProp Σ).
  Context {HP : Fractional P}.

  Global Instance biabd_alloc_none q φ mq :
    CmraSubtract 1%Qp q φ mq →
    HINT ε₁ ✱ [- ; ⌜φ⌝] ⊫ [bupd] γ; no_tokens P γ q ✱ [default emp (q' ← mq; Some $ no_tokens P γ q')].
  Proof. apply: tokenizable.biabd_alloc_none. Qed.

  Global Instance biabd_alloc_some :
    HINT ε₁ ✱ [ - ; P 1] ⊫ [bupd] γ; token_counter P γ 1 ✱ [token P γ].
  Proof. apply: tokenizable.biabd_alloc_some. Qed.

  Global Instance biabd_create_token γ p1 p2 :
    SolveSepSideCondition (p2 = Pos.succ p1) →
    HINT token_counter P γ p1 ✱ [- ; emp] ⊫ [bupd]; token_counter P γ p2 ✱ [token P γ].
  Proof. apply: tokenizable.biabd_create_token. Qed.

  Global Instance biabd_create_first_token γ p :
    SolveSepSideCondition (p = 1%positive) →
    HINT no_tokens P γ 1 ✱ [- ; P 1] ⊫ [bupd]; token_counter P γ p ✱ [token P γ].
  Proof. apply: tokenizable.biabd_create_first_token. Qed.

  Global Instance biabd_delete_token γ p1 p2 :
    SolveSepSideCondition (1 < p1)%positive →
    SolveSepSideCondition (p2 = Pos.pred p1) →
    HINT token_counter P γ p1 ✱ [- ; token P γ] ⊫ [bupd]; token_counter P γ p2 ✱ [emp].
  Proof. apply: tokenizable.biabd_delete_token. Qed.

  Global Instance biabd_delete_last_token γ q mq φ p :
    SolveSepSideCondition (p = 1)%positive →
    CmraSubtract 1%Qp q φ mq →
    HINT token_counter P γ p ✱ [- ; token P γ ∗ ⌜φ⌝] ⊫ [bupd]; no_tokens P γ q ✱ [P 1 ∗ default emp (q' ← mq; Some $ no_tokens P γ q')].
  Proof. apply: tokenizable.biabd_delete_last_token. Qed.

  Global Instance biabd_delete_last_token_to_req γ p :
    SolveSepSideCondition (p = 1%positive) →
    HINT token_counter P γ p ✱ [- ; token P γ] ⊫ [bupd]; P 1 ✱ [no_tokens P γ 1].
  Proof. apply: tokenizable.biabd_delete_last_token_to_req. Qed.

  Global Instance no_tokens_split γ q1 q2 mq :
    FracSub q1 q2 mq →
    HINT no_tokens P γ q1 ✱ [- ; True] ⊫ [id]; no_tokens P γ q2 ✱ [match mq with | Some q => no_tokens P γ q | None => True end].
  Proof. apply: tokenizable.no_tokens_split. Qed.

  Fixpoint token_iter n γ : iProp Σ :=
    match n with
    | O => emp
    | S n => token P γ ∗ token_iter n γ
    end.

  Lemma token_iter_eq γ n : token_iter n γ = tokenizable.tokenizable.token_iter (fractional_payload_token_counter P) n γ.
  Proof.
    induction n as [| n IH] => //=.
    by rewrite IH.
  Qed.

  Global Instance merge_token_no_tokens γ q :
    MergableConsume (▷ token P γ) (λ p Pin Pout,
      TCAnd (TCEq Pin (no_tokens P γ q))
            (TCEq Pout (◇ False)))%I | 11.
  Proof. apply: tokenizable.merge_token_no_tokens. Qed.

  Global Instance merge_token_no_tokens_now γ q :
    MergableConsume (token P γ) (λ p Pin Pout,
      TCAnd (TCEq Pin (no_tokens P γ q))
            (TCEq Pout (◇ False)))%I | 11.
  Proof. apply: tokenizable.merge_token_no_tokens_now. Qed.

  Global Instance merge_no_tokens_token_now γ q :
    MergableConsume (no_tokens P γ q) (λ p Pin Pout,
      TCAnd (TCEq Pin (token P γ))
            (TCEq Pout (◇ False)))%I | 10.
  Proof. apply: tokenizable.merge_no_tokens_token_now. Qed.

  Global Instance merge_no_tokens_token γ q :
    MergableConsume (no_tokens P γ q) (λ p Pin Pout,
      TCAnd (TCEq Pin (▷ token P γ))
            (TCEq Pout (◇ False)))%I | 10.
  Proof. apply: tokenizable.merge_no_tokens_token. Qed.

  Global Instance merge_no_tokens_token_later_counter γ q pos:
    MergableConsume (no_tokens P γ q) (λ p Pin Pout,
      TCAnd (TCEq Pin (▷ token_counter P γ pos))
            (TCEq Pout (◇ False)))%I | 11.
  Proof. apply: tokenizable.merge_no_tokens_token_later_counter. Qed.

  Global Instance merge_no_tokens_token_counter γ q pos:
    MergableConsume (no_tokens P γ q) (λ p Pin Pout,
      TCAnd (TCEq Pin (token_counter P γ pos))
            (TCEq Pout (◇ False)))%I | 12.
  Proof. apply: tokenizable.merge_no_tokens_token_counter. Qed.

  Global Instance merge_no_tokens γ q q1 q2 :
    MergableConsume (no_tokens P γ q1) (λ p Pin Pout,
      TCAnd (TCEq Pin (no_tokens P γ q2)) $
      TCAnd (IsOp q q1 q2) 
            (TCEq Pout (no_tokens P γ q ∗ ⌜q ≤ 1⌝%Qp)))%I | 20.
  Proof. apply: tokenizable.merge_no_tokens. Qed.

  Global Instance merge_token_counter_no_tokens γ q pos:
    MergableConsume (token_counter P γ pos) (λ p Pin Pout,
      TCAnd (TCEq Pin (no_tokens P γ q))
            (TCEq Pout (◇ False)))%I | 11.
  Proof. apply: tokenizable.merge_token_counter_no_tokens. Qed.

  Global Instance merge_token_counter_no_tokens_later γ q pos:
    MergableConsume (▷ token_counter P γ pos)%I (λ p Pin Pout,
      TCAnd (TCEq Pin (no_tokens P γ q))
            (TCEq Pout (◇ False)))%I | 11.
  Proof. apply: tokenizable.merge_token_counter_no_tokens_later. Qed.

  Global Instance no_tokens_fractional γ : Fractional (no_tokens P γ).
  Proof. iSolveTC. Qed.

  Global Instance no_tokens_timeless γ q : Timeless (no_tokens P γ q).
  Proof. iSolveTC. Qed.

  Global Instance token_into_exist γ : 
    (* this one should currently be manually enabled with Existing Instance token_into_exist *)
    AtomIntoExist (token P γ) (λ q : Qp, P q ∗ ◇ (P q -∗ token P γ))%I.
  Proof. 
    rewrite /AtomIntoExist /IntoExistCareful2 /IntoExistCareful. iStepsS.
    iApply bi.except_0_intro. iStepsS.
  Qed.

End automation.

Global Opaque token_counter.
Global Opaque no_tokens.
Global Opaque token.

(* this is missing and sometimes causes very slow TC resolution.
    TODO: merge into Iris? *)
Global Hint Mode Fractional + + : typeclass_instances.

Section token_counter_extra.
  Context `{!fracTokenG Σ}.
  Context (P : Qp → iProp Σ).
  Context {HP : Fractional P}.

  Global Instance biabd_alloc_multiple (p : positive) :
    HINT ε₁ ✱ [- ; P 1] ⊫ [bupd] γ; token_counter P γ p ✱ [token_iter P (Pos.to_nat p) γ] | 500.
  Proof.
    rewrite /BiAbd /= empty_hyp_last_eq left_id.
    rewrite -{1}(Pos2Nat.id p).
    assert (0 < Pos.to_nat p) by lia. revert H.
    generalize (Pos.to_nat p) => n {p}.
    induction n as [|[|n] IHn]; first lia.
    - clear IHn => _. cbn. iStepsS.
    - intros Hn.
      rewrite IHn; last by lia.
      iStepsS.
  Qed.

End token_counter_extra.



