From iris.proofmode Require Import base coq_tactics reduction tactics.
From diaframe Require Import env_utils util_classes solve_defs tele_utils.
From diaframe.steps Require Import namer_tacs pure_solver.
From iris.bi Require Import bi telescopes.

Import bi.

(* This file contains 'case 2' of Diaframe's proof search strategy: introducing wands.
   In addition to the paper, this also includes using MergableConsume and
   MergablePersist instances to detect compatibilities between hypotheses in the context *)

Section mergable.
  Context {PROP : bi}.
  Implicit Types P : PROP.

  Lemma merge_and_consume P1 C p P2 P3 R:
    MergableConsume P1 C →
    C p P2 P3 →
    □?p P2 ∗ (P3 -∗ R) ⊢ P1 -∗ R.
  Proof.
    move => Hmerge HC /=.
    iIntros "[HP2 HPR] HP1".
    iApply "HPR".
    iApply mergable_consume; [done | iFrame].
  Qed.

  Lemma merge_no_consume P1 C p P2 P3 R :
    MergablePersist P1 C →
    C p P2 P3 →
    □?p P2 ∗ (□?p P2 ∗ P1 -∗ □ P3 -∗ R) ⊢ P1 -∗ R.
  Proof.
    move => Hmerge HC /=.
    iIntros "[HP2 HPR] HP1".
    iAssert (<pers> P3)%I with "[HP1 HP2]" as "#HP3".
    { iApply mergable_persist; [done | iFrame]. }
    by iApply ("HPR" with "[$HP1 $HP2]").
  Qed.

End mergable.


Section mergable_env.
  Context {PROP : bi}.

  (* MergablePersist will only detect one hypothesis with a persistent compatibility.
     Sometimes there are more. The [MergableEnv] typeclass ensures all of them are found *)
  Class MergableEnv (Δ : envs PROP) (P : PROP) (M : option PROP) :=
    mergable_env : (of_envs Δ) ∗ P ⊢ <pers> default emp M.

  Instance mergable_env_base Δ P : 
    MergableEnv Δ P None | 999.
  Proof. rewrite /MergableEnv; eauto. Qed.

  (* This instance is a bit crappy, but needed since FindInExtendedContext will keep finding
     [εᵢ] instances, while we only want to do that once. *)
  Global Instance mergable_env_base2 C0 C1 P P0 mP0 P1 mP1 Δ :
    TCOr (TCAnd (MergablePersist P C0) $
          TCAnd (C0 false (ε₀)%I P0) $
                 TCEq mP0 (Some P0)) 
         (TCAnd (TCEq C0 (λ _ _ _, True)) $
          TCAnd (TCEq P0 bi_emp) $
                 TCEq mP0 None) →    
    TCOr (TCAnd (MergablePersist P C1) $
          TCAnd (C1 false (ε₁)%I P1) $
                 TCEq mP1 (Some P1)) 
         (TCAnd (TCEq C1 (λ _ _ _, True)) $
          TCAnd (TCEq P1 bi_emp) $
                 TCEq mP1 None) →
    MergableEnv Δ P (match mP1 with 
                     | Some H1 => Some $ match mP0 with | Some H0 => H0 ∧ H1 | _ => H1 end
                     | None => mP0
                     end)%I | 999.
  Proof.
    rewrite /MergableEnv.
    move => [[HPC0 [/HPC0 /= HC0 ->]]|[_ [_ ->]]] [[HPC1 [/HPC1 /= HC1 ->]]|[_ [_ ->]]] /=.
    - iIntros "[_ HP]". iSplit.
      * iApply HC0. iFrame. rewrite empty_hyp_first_eq //.
      * iApply HC1. iFrame. rewrite empty_hyp_last_eq //.
    - iIntros "[_ HP]".
      iApply HC0. iFrame. rewrite empty_hyp_first_eq //.
    - iIntros "[_ HP]".
      iApply HC1. iFrame. rewrite empty_hyp_last_eq //.
    - eauto.
  Qed.

  Global Instance mergable_env_cons Δ P C i p P2 P3 M :
    MergablePersist P C →
    FindInContext Δ (λ p P2, C p P2 P3) i p P2 →
    MergableEnv (envs_delete true i p Δ) P M →
    MergableEnv Δ P (Some $ match M with | Some M' => M' ∧ P3 | _ => P3 end)%I | 100.
  Proof.
    rewrite /MergablePersist /MergableEnv => HCP HΔ.
    apply (fic_satisfies) in HΔ as HC. simpl in HC.
    apply (findincontext_spec true) in HΔ.
    rewrite HΔ => {HΔ}. set (Δ' := envs_delete true i p Δ).
    rewrite -assoc => HΔ.
    iIntros "(HP2 & HΔ & HP)". cbn.
    destruct M as [M'|]; first iSplit.
    - iClear "HP2". iApply HΔ. by iFrame.
    - iClear "HΔ". iApply HCP => //. by iFrame.
    - iClear "HΔ". iApply HCP => //. by iFrame.
  Qed.
End mergable_env.


Section lemmas.
  Context {PROP : bi}.

  Implicit Types P Q : PROP.

  Lemma introduce_hyp_premise_sep P P1 P2 Q :
    IntoSepCareful P P1 P2 →
    IntroduceHyp P1 (IntroduceHyp P2 Q) ⊢ IntroduceHyp P Q.
  Proof.
    unseal_diaframe; rewrite /IntoSepCareful /= => ->.
    iIntros "HPQ [HP1 HP2]".
    by iApply ("HPQ" with "HP1").
  Qed.

  Lemma introduce_hyp_premise_tele_exist {TT} P R Q :
    IntoTExist P TT R →
    IntroduceVars (TT := TT) $ tele_map (λ R, IntroduceHyp R Q) R ⊢ IntroduceHyp P Q.
  Proof.
    unseal_diaframe; rewrite /IntoTExist /= => ->.
    iIntros "HRQ".
    iDestruct 1 as (tt) "HR".
    iSpecialize ("HRQ" $! tt).
    rewrite tele_map_app.
    by iApply "HRQ".
  Qed.

  (* Currently, IntoTExist cannot handle dependent exists. 
  For IntroduceHyp, we can overcome this by introducing the exists one by one: *)

  Lemma introduce_hyp_premise_exist (P Q : PROP) {A : Type} (R : A → PROP) :
    IntoExistCareful P R →
    IntroduceVars (TT := [tele_pair A]) (λ a, IntroduceHyp (R a) Q) ⊢ IntroduceHyp P Q.
  Proof.
    unseal_diaframe; rewrite /IntoExistCareful /= => ->.
    iIntros "HRQ".
    iDestruct 1 as (r) "HR".
    by iApply "HRQ".
  Qed.

  Lemma introduce_hyp_premise_or P P1 P2 Q : 
    IntoOr P P1 P2 →
    IntroduceHyp P1 Q ∧ IntroduceHyp P2 Q ⊢ IntroduceHyp P Q.
  Proof. 
    unseal_diaframe; rewrite /IntoOr /= => ->. 
    iIntros "HPQ [HP1|HP2]".
    - iDestruct "HPQ" as "[HPQ1 _ ]".
      by iApply "HPQ1".
    - iDestruct "HPQ" as "[_ HPQ2]".
      by iApply "HPQ2".
  Qed.

  Lemma introduce_hyp_mergable_consume p P C Q R S :
    MergableConsume Q C →
    C p P R →
    □?p P ∗ IntroduceHyp R S ⊢ IntroduceHyp Q S.
  Proof.
    unseal_diaframe => Hmerge HC.
    eapply merge_and_consume => //.
  Qed.

  Lemma introduce_hyp_mergable_no_consume p P C Q R S :
    MergablePersist Q C →
    C p P R →
    □?p P ∗ (□?p P ∗ Q -∗ IntroduceHyp (□ R) S) ⊢ IntroduceHyp Q S.
  Proof.
    unseal_diaframe => /= Hmerge HC.
    eapply merge_no_consume => //.
  Qed.

  Lemma introduce_hyp_premise_intuit P P' Q :
    IntoPersistent false P P' →
    Affine P →
    (□ P' -∗ Q) ⊢ IntroduceHyp P Q.
  Proof.
    unseal_diaframe; rewrite /IntoPersistent /= => -.
    iIntros (HPP' HP) "HP'Q HP".
    iApply "HP'Q".
    iStopProof.
    rewrite /bi_intuitionistically.
    by apply affinely_intro.
  Qed.

  Lemma introduce_hyp_drop_modal P M P' C Q :
    IntoModal false P M P' →
    DropModal M Q C →
    C P' →
    IntroduceHyp P' Q ⊢ IntroduceHyp P Q.
  Proof.
    unseal_diaframe; rewrite /IntoModal /DropModal /= => -> HQ /HQ //.
  Qed.

End lemmas.

Section coq_tactics.
  Context {PROP : bi}.

  Implicit Types P Q : PROP.

  Lemma tac_introduce_hyp_premise_sep Δ P P1 P2 Q :
    IntoSepCareful P P1 P2 →
    envs_entails Δ $ IntroduceHyp P1 (IntroduceHyp P2 Q) →
    envs_entails Δ $ IntroduceHyp P Q.
  Proof.
    rewrite envs_entails_unseal => HΔΔ' ->.
    apply introduce_hyp_premise_sep, _.
  Qed.

  Lemma tac_introduce_hyp_premise_exist Δ P {A} (R : A → PROP) Q:
    IntoExistCareful P R →
    envs_entails Δ $ IntroduceVars (TT := [tele_pair A]) (λ a, IntroduceHyp (R a) Q) →
    envs_entails Δ $ IntroduceHyp P Q.
  Proof.
    rewrite envs_entails_unseal => HPR ->.
    by eapply introduce_hyp_premise_exist.
  Qed.

  Lemma tac_introduce_hyp_premise_or Δ P P1 P2 Q :
    IntoOr P P1 P2 →
    envs_entails Δ $ IntroduceHyp P1 Q →
    envs_entails Δ $ IntroduceHyp P2 Q →
    envs_entails Δ $ IntroduceHyp P Q.
  Proof.
    rewrite envs_entails_unseal => HP HΔ1 HΔ2.
    rewrite -introduce_hyp_premise_or.
    by apply and_intro.
  Qed.

  Lemma tac_introduce_hyp_premise_modal Δ P M P' C Q :
    IntoModal false P M P' →
    DropModal M Q C →
    C P' →
    envs_entails Δ $ IntroduceHyp P' Q →
    envs_entails Δ $ IntroduceHyp P Q.
  Proof.
    rewrite envs_entails_unseal => HP HQC HCP ->.
    by eapply introduce_hyp_drop_modal.
  Qed.

  Lemma tac_introduce_hyp_merge_consume Δ P1 R C mi p P2 P3 :
    TCAnd (MergableConsume P1 C) (FindInExtendedContext Δ (λ p P2, C p P2 P3) mi p P2) →
    envs_entails (envs_option_delete true mi p Δ) $ IntroduceHyp P3 R →
    envs_entails Δ $ IntroduceHyp P1 R.
  Proof.
    rewrite envs_entails_unseal => [[HP1 HP2]] HΔ.
    rewrite (findinextcontext_spec true) HΔ.
    eapply introduce_hyp_mergable_consume => //.
    by eapply ficext_satisfies in HP2.
  Qed.

  Lemma tac_introduce_hyp_merge_persist j Δ P1 a P1' R C mi p P2 P3 :
    TCAnd (MergablePersist P1 C) (FindInExtendedContext Δ (λ p P2, C p P2 P3) mi p P2) →
    TCIf (TCAnd (IntoPersistent false P1 P1') (Affine P1))
        (TCEq a true) (TCAnd (TCEq a false) (TCEq P1' P1)) →
    match envs_add_fresh a j P1' Δ with
    | Some Δ' => envs_entails Δ' $ IntroduceHyp (□ P3)%I R
    | _ => False
    end →
    envs_entails Δ $ IntroduceHyp P1 R.
  Proof.
    unseal_diaframe; rewrite envs_entails_unseal => [[HP1 HP2]] HP1p' HΔR.
    destruct (envs_add_fresh a j P1' Δ) as [Δ'| ] eqn:HΔ''; last done.
    iIntros "HΔ' HP1".
    assert (P1 ⊢ □?a P1') as HP1_ent.
    { revert HP1p'. case => [[HP1P1' HP1a] ->| [-> ->] ] //=.
      unfold bi_intuitionistically.
      rewrite -HP1P1' /=. unfold bi_affinely. iIntros "H"; iSplit => //. }
    iAssert (□ P3)%I as "#HP3"; last first. (* 'iEnough' *)
    { rewrite (envs_add_fresh_sound Δ) //.
      rewrite HP1_ent.
      iSpecialize ("HΔ'" with "HP1"). 
      by iApply (HΔR with "HΔ'"). }
    rewrite (findinextcontext_spec true).
    iDestruct "HΔ'" as "[HP2 _]".
    iCombine "HP1 HP2" as "HP".
    rewrite mergable_persist; last first.
    { by eapply ficext_satisfies in HP2. }
    done.
  Qed.

  Lemma tac_introduce_hyp_merge_env j Δ P1 a P1' R mP3 :
    MergableEnv Δ P1 mP3 →
    TCIf (TCAnd (IntoPersistent false P1 P1') (Affine P1))
        (TCEq a true) (TCAnd (TCEq a false) (TCEq P1' P1)) →
    match envs_add_fresh a j P1' Δ with
    | Some Δ' => 
      match mP3 with
      | None => envs_entails Δ' R
      | Some P3 =>
        envs_entails Δ' $ IntroduceHyp (□ P3)%I R
      end
    | _ => False
    end →
    envs_entails Δ $ IntroduceHyp P1 R.
  Proof.
    unseal_diaframe; rewrite envs_entails_unseal => HΔ HP1p' HΔR.
    destruct (envs_add_fresh a j P1' Δ) as [Δ'| ] eqn:HΔ''; last done.
    iIntros "HΔ' HP1".
    assert (P1 ⊢ □?a P1') as HP1_ent.
    { revert HP1p'. case => [[HP1P1' HP1a] ->| [-> ->] ] //=.
      unfold bi_intuitionistically.
      rewrite -HP1P1' /=. unfold bi_affinely. iIntros "H"; iSplit => //. }
    destruct mP3 as [P3|]; last first.
    { rewrite envs_add_fresh_sound // -HΔR. rewrite -HP1_ent. by iApply "HΔ'". }
    iAssert (□ P3)%I as "#HP3"; last first. (* 'iEnough' *)
    { rewrite (envs_add_fresh_sound Δ) //.
      rewrite -HP1_ent.
      iSpecialize ("HΔ'" with "HP1"). 
      by iApply (HΔR with "HΔ'"). }
    iCombine "HΔ' HP1" as "H".
    rewrite HΔ.
    eauto.
  Qed.

  (* TODO: move these to solve_instances? *)
  Global Instance mergable_persist_from_box P C :
    MergablePersist P C →
    MergablePersist (□ P)%I C | 5.
  Proof.
    rewrite /MergablePersist => HPC p P2 P3 /HPC <-.
    iIntros "[#$ $]".
  Qed.
  Global Instance mergable_consume_from_box P C :
    MergableConsume P C →
    MergableConsume (□ P)%I C | 5.
  Proof.
    rewrite /MergableConsume => HPC p P2 P3 /HPC <-.
    iIntros "[#$ $]".
  Qed.

  Lemma tac_introduce_hyp_not_mergable j Δ P1 a P1' R : 
    TCIf (TCAnd (IntoPersistent false P1 P1') (Affine P1))
        (TCEq a true) (TCAnd (TCEq a false) (TCEq P1' P1)) →
    match envs_add_fresh a j P1' Δ with
    | Some Δ' => envs_entails Δ' $ R
    | _ => False
    end →
    envs_entails Δ $ IntroduceHyp P1 R.
  Proof.
    unseal_diaframe; rewrite envs_entails_unseal => HP1p' HΔR.
    destruct (envs_add_fresh a j P1' Δ) as [Δ'| ] eqn:HΔ''; last done.
    rewrite (envs_add_fresh_sound Δ) // HΔR.
    apply bi.wand_mono => //.
    revert HP1p'. case => [[HP1P1' HP1a] ->| [-> ->] ] //=.
    unfold bi_intuitionistically.
    rewrite -HP1P1' /=. unfold bi_affinely. iIntros "H"; iSplit => //.
  Qed.

  Lemma tac_introduce_hyp_premise_pure Δ P φ Q ba :
    IntoPure P φ →
    TCIf (Affine P) (TCEq ba true) (TCEq ba false) →
    match (if ba then Some Δ else envs_add_fresh false None True%I Δ) with
    | Some Δ' => φ → envs_entails Δ' Q
    | _ => False
    end →
    envs_entails Δ $ IntroduceHyp P Q.
  Proof.
    rewrite envs_entails_unseal => HPφ HP.
    destruct HP as [HP ->| ->].
    - assert (IntoPersistent false P ⌜φ⌝).
      { rewrite /IntoPersistent /= HPφ.
        by rewrite -(persistently_if_pure true φ) /= persistently_idemp. }
      rewrite -introduce_hyp_premise_intuit => HΔφ.
      iIntros "HΔ %".
      by iApply HΔφ.
    - unseal_diaframe.
      destruct (envs_add_fresh false None True Δ) as [Δ'| ] eqn:HΔ''; last done.
      rewrite HPφ => Hφ.
      apply bi.wand_intro_l.
      apply bi.wand_elim_l', bi.pure_elim' => /Hφ.
      apply envs_add_fresh_sound in HΔ''.
      rewrite HΔ'' /= => <-.
      apply bi.wand_intro_r.
      rewrite bi.wand_elim_r //.
  Qed.

  Lemma tac_introduce_hyp_premise_emp Δ Q :
    envs_entails Δ Q →
    envs_entails Δ $ IntroduceHyp emp%I Q.
  Proof. 
    unseal_diaframe; rewrite envs_entails_unseal => ->.
    by rewrite left_id. 
  Qed.

End coq_tactics.


(* Now comes some infrastructure to revert hypotheses when a variable occurring in that hypothesis
   has been used for a rewrite *)
Fixpoint revert_idents {PROP : bi} (js : list ident) Δ G : option (envs PROP * PROP) :=
  match js with
  | [] => Some (Δ, G)
  | j :: js =>
    '(Δ', G') ← revert_idents js Δ G;
    '(p, H) ← envs_lookup j Δ';
    Some (envs_delete true j p Δ', IntroduceHyp H G')
  end.

Lemma revert_idents_sound {PROP : bi} (js : list ident) Δ (G : PROP) :
  match revert_idents js Δ G with
  | Some (Δ', G') => envs_entails Δ' G'
  | _ => False
  end →
  envs_entails Δ G.
Proof.
  rewrite envs_entails_unseal.
  revert Δ G.
  induction js as [|j js] => //.
  move => Δ G. cbn.
  destruct (revert_idents _ _ _) as [[Δ' G']|] eqn:HΔG => //=.
  destruct (envs_lookup _ _) as [[p P]|] eqn:HjΔ => //=.
  move => HΔP.
  apply IHjs. rewrite HΔG.
  rewrite envs_lookup_sound //.
  apply bi.wand_elim_r'.
  rewrite HΔP. unseal_diaframe.
  by rewrite bi.intuitionistically_if_elim.
Qed.

(* some ltacs to collect hypotheses which should be reintroduced *)
Ltac collect_env_occs x Γi acci :=
  let rec mthd Γ acc :=
    lazymatch Γ with
    | Enil => acc
    | Esnoc ?Γ' ?i ?H =>
      lazymatch H with
      | context [x] => 
        let result := mthd Γ' acc in
        constr:(i::result)
      | _ => mthd Γ' acc
      end
    end
  in mthd Γi acci.

Ltac collect_envs_occs x Δ :=
  let init := constr:(@nil ident) in
  let Γs := eval cbn in (env_spatial Δ) in
  let result_s := collect_env_occs x Γs init in
  let Γp := eval cbn in (env_intuitionistic Δ) in
  collect_env_occs x Γp result_s.

Ltac my_subst x :=
  assert_succeeds (subst x);
  lazymatch goal with
  | |- envs_entails ?Δ ?G =>
    let result := collect_envs_occs x Δ in
    refine (revert_idents_sound result Δ G _); simpl;
    subst x
  end.

Lemma do_simplify_hyp (φin φout : Prop) (Hφ : φin) (G : Prop) : SimplifyPureHyp φin φout → (φout → G) → G.
Proof. rewrite /SimplifyPureHyp. eauto. Qed.

(* adapted from stdpp's simplify_eq *)
Ltac simplify_hyp_eq_base cont :=
  match goal with
  | |- ?Htype → ?G => 
    let H := fresh "H" in
    intro H;
    repeat
    match type of H with
    | _ ∧ _ => case H; try clear H; cont; cont
    | ∃ _, _ => case H; try clear H; intros ?; cont
    | _ ≠ _ => by case H; try clear H
    | _ = _ → False => by case H; try clear H
    | ?x = _ => my_subst x
    | _ = ?x => my_subst x
    | _ = _ => discriminate H
    | _ ≡ _ => apply leibniz_equiv in H
    | ?f _ = ?f _ => apply (inj f) in H
    | ?f _ _ = ?f _ _ => apply (inj2 f) in H; destruct H
      (* before [injection] to circumvent bug #2939 in some situations *)
    | ?f _ = ?f _ => progress injection H as H
      (* first hyp will be named [H], subsequent hyps will be given fresh names *)
    | ?f _ _ = ?f _ _ => progress injection H as H
    | ?f _ _ _ = ?f _ _ _ => progress injection H as H
    | ?f _ _ _ _ = ?f _ _ _ _ => progress injection H as H
    | ?f _ _ _ _ _ = ?f _ _ _ _ _ => progress injection H as H
    | ?f _ _ _ _ _ _ = ?f _ _ _ _ _ _ => progress injection H as H
    | ?x = ?x => clear H
    | @existT ?A _ _ _ = existT _ _ =>
       apply (Eqdep_dec.inj_pair2_eq_dec _ (decide_rel (=@{A}))) in H
    | _ => notypeclasses refine (do_simplify_hyp _ _ H _ _ _); first iSolveTC; try clear H; cont
    end
  end.

Ltac simplify_hyp_eq := simplify_hyp_eq_base ltac:(idtac; simplify_hyp_eq).

Ltac simplifySolveSepPureHyp := 
  lazymatch goal with
  | |- ?φ → ?P => 
    first
    [ let H := fresh "H" in assert (¬φ) as H by trySolvePure; by move => /H
    (* tries to prove ¬φ, to eliminate contradictory cases *)
    | lazymatch φ with
      | True => intros _
      | ?a = ?a => intros _
      | _ => simplify_hyp_eq
      end]
  end.

Ltac introduceHypEmp := 
  notypeclasses refine (tac_introduce_hyp_premise_emp _ _ _ ).

Ltac introduceHypPure := 
  notypeclasses refine (tac_introduce_hyp_premise_pure _ _ _ _ _ _ _ _); 
  [ iSolveTC | iSolveTC | simpl; simplifySolveSepPureHyp ].

Ltac introduceHypSep := 
  notypeclasses refine (tac_introduce_hyp_premise_sep _ _ _ _ _ _ _); [ iSolveTC | ].

Ltac introduceHypExist := 
  notypeclasses refine (tac_introduce_hyp_premise_exist _ _ _ _ _ _); [ iSolveTC | simpl].

Ltac introduceHypOr :=
  notypeclasses refine (tac_introduce_hyp_premise_or _ _ _ _ _ _ _ _ ); [ iSolveTC | simpl | simpl].

Ltac introduceHypStep namer_tac :=
  lazymatch goal with
  | |- envs_entails ?Δ (IntroduceHyp ?P ?G) =>
    lazymatch P with
    | bi_emp => introduceHypEmp
    | bi_sep _ _ => introduceHypSep
    | bi_exist _ => introduceHypExist
    | bi_or _ _ => introduceHypOr
    | bi_pure ?φ => 
      lazymatch φ with
      | ?φl ∧ ?φr => introduceHypSep
      | ?φl ∨ ?φr => introduceHypOr
      | ex _ => introduceHypExist
      | _ => introduceHypPure
      end
    | _ =>
      first
      [introduceHypEmp
      |introduceHypSep
      |introduceHypExist
      |introduceHypOr
      |notypeclasses refine (tac_introduce_hyp_premise_modal _ _ _ _ _ _ _ _ _ _); 
        [iSolveTC | iSolveTC | simpl; iSolveTC | ]
      (* rules below this do actual introduction of separation logic premises *)
      |introduceHypPure
      |notypeclasses refine (tac_introduce_hyp_merge_consume _ _ _ _ _ _ _ _ _ _); 
        [iSolveTC | register_delete namer_tac; simpl ]
      | let j := namer_tac NameTac.GetName Δ P in 
        notypeclasses refine (tac_introduce_hyp_merge_env j _ _ _ _ _ _ _ _ _);
        [iSolveTC | iSolveTC | simpl]
      (* MergeEnv covers the case where no merging needs to be done, so this rule is no longer necessary.
         However, it would be faster than MergeEnv for hypotheses that dont need merging.
      |notypeclasses refine (tac_introduce_hyp_not_mergable j _ _ _ _ _ _ _);
        [iSolveTC | simpl ] ] *)
      ]
    end
  end.

