From iris.proofmode Require Import base coq_tactics reduction tactics.
From diaframe Require Import util_classes.


(* When introducing hypothesis, Diaframe needs to give them a name.
   One option would be to always give these an anonymous name. 
   This is not the most user-friendly option in the case of partial proofs, but possibly more performant.
   To make choosing names modular, this file defines a 'namer_tac': a tactic which should be be able to provide
   a name for hypothesis when asked, and which will also be called when hypotheses get deleted.

   The current default namer will name hypotheses "Hn" with n a number. This is not the most consistent naming and
   could be improved.
*)


Module NameTac.
  Inductive NameTac :=
    | GetName
    | RegisterDelete.
End NameTac.

Ltac register_delete namer_tac :=
  lazymatch goal with
  | |- envs_entails (envs_delete ?rp ?i ?p ?Δ) _ =>
    let real_delete := eval cbn in (orb rp $ negb p) in
    lazymatch real_delete with
    | true => namer_tac NameTac.RegisterDelete i
    | _ => idtac
    end
  | |- envs_entails (envs_option_delete ?rp ?mi ?p ?Δ) _ =>
    lazymatch mi with
    | Some ?i =>
      let real_delete := eval cbn in (orb rp $ negb p) in
      lazymatch real_delete with
      | true => namer_tac NameTac.RegisterDelete i
      | _ => idtac
      end
    | None => idtac
    end
  | |- ?P => fail 5 "Register delete was called on incompatible goal "P" BAD!"
  end.

(* this namer always gives hypothesis an anonymous name *)
Ltac anon_namer arg :=
  lazymatch arg with
  | NameTac.GetName => 
      fun Δ P => constr:(@None ident)
  | NameTac.RegisterDelete => 
      fun i => idtac
  end.

Ltac get_fresh_name Δ base' append :=
  let rec fresh_name base :=
    let name_in_env := eval cbv in (existsb (fun i => ident_beq i base) (envs_dom Δ)) in
    match constr:(name_in_env) with
    | true => 
      let new_name := eval cbv in (base +:+ append) in
      fresh_name new_name
    | false => base
    end 
  in fresh_name base'.

(* this namer names new hypothesis 'Hn', where n is 'env_counter Δ'. If not fresh, it appends apostrophes until fresh *)
Ltac counter_namer arg := 
  lazymatch arg with
  | NameTac.GetName => 
      fun Δ P => (* quite crappy, but at least it gives a fresh usable name *)
        let base := eval cbv in (String.append "H" (pretty.pretty_positive (env_counter Δ))) in
        let ident := get_fresh_name Δ base "'" in
        constr:(Some $ INamed ident)
  | NameTac.RegisterDelete => 
      fun i => idtac
  end.

(* this namer names new hypothesis 'Hn', where 'n ≥ env_counter Δ', and '∀ k, env_counter Δ ≤ k < n → Hk not fresh' *)
Ltac counter_namer_v2 arg := 
  lazymatch arg with
  | NameTac.GetName => 
      fun Δ P => (* does someting nice than just appending apostrophes. *)
        let rec first_fresh num :=
          let base := eval cbv in (String.append "H" (pretty.pretty_positive num)) in
          let name_in_env := eval cbv in (existsb (fun i => ident_beq i base) (envs_dom Δ)) in
          match constr:(name_in_env) with
          | true => 
            let new_num := eval cbv in (Pos.succ num) in
            first_fresh new_num
          | false => base
          end in
        let start_num := eval cbn in (env_counter Δ) in
        let ident := first_fresh start_num in
        constr:(Some $ INamed ident)
  | NameTac.RegisterDelete => 
      fun i => idtac
  end.

Ltac default_namer := counter_namer_v2.