From iris.proofmode Require Import base coq_tactics reduction tactics.
From diaframe Require Import env_utils util_classes solve_defs.
From diaframe.steps Require Import namer_tacs pure_solver solve_sep transform introduce_hyp introduce_var solve_one.
From iris.bi Require Import bi telescopes.


(* Apart from iStepS, Diaframe also defines a (more fine-grained (😉) ) solveStep tactic,
   which executes precisely one instruction of Diaframe's proof search strategy.

   If a hint you expect to be applied is not applied or found, this file comes with a 'solveStep with H' tactic.
   The next goals will be:
   - prove a hint from hypothesis H to the current focused atom
   - continuation goal after applying this hint.
*)

(* We need to do some juggling with evars, which the next few tactics are for *)
Ltac hide_evars_adv next_tac name_hook := 
  match goal with
  | |- context [?arg] =>
    is_evar arg;
    let evar_name := fresh "x" in
    set (evar_name := arg);
    try (hide_evars_adv next_tac name_hook); 
    (* it is important that failure of next_tac will not cause evars to be unfolded *)
    name_hook evar_name
  | |- _ => next_tac
  end.

Ltac unhider evar_name :=
  unfold evar_name; clear evar_name.

Ltac noop_tac evar_name := idtac.

Ltac hide_evars_then next_tac :=
  hide_evars_adv next_tac unhider.

Tactic Notation "hide_evars" "then" tactic(next) "then" "unhide" :=
  hide_evars_then next.

Tactic Notation "hide_evars" := hide_evars_adv idtac noop_tac.

(* Now, we define the tactic that tries to solve pure goals.
   It fails if the pure goal has to be solved now, but can't (since some evar needs to be instantiated).
   It shelves the pure goal if it cant be solved and doesnt have to be solved now *)
Ltac trySolvePurePreSplit φ R :=
  (* fancy pure stuff from RefinedC should go here *)
  first (* check if we are not already unprovable. If we are, fail hard. *)
  [ match φ with
    | @eq _ ?a ?b => is_evar a; fail 1 "dont try negating evar equalities"
    | _ => idtac
    end; 
    assert (¬ φ) as _ by (hide_evars then (idtac; trySolvePure) then unhide); 
    (* if idtac is not present, trySolvePure will run immediately...? *)
    (* error level 4 is carefully chosen so that the failure stops iStepsS/iSmash, but does not make it give an error *)
    fail 4 "Current pure goal is provably untrue! Something went terribly wrong!"
  | has_evar φ; 
    ((split; first (solve [trySolvePure])) ||
    (* if there is an evar, we need to solve it now *)
    (* somehow prepending idtac changes the semantics and makes it work like intentioned...?*)
    (idtac; fail 1 "Pure condition"φ"needs to be proved now!"))
  (* final option: solve it now, and if unsuccesful, shelve it *)
   | split; [first [solve [trySolvePure] | shelve] | ] ].

Ltac solveStepNT namer_tac :=
  (simplify_eq; subst) ||
  lazymatch goal with
  | |- envs_entails _ ?R =>
    lazymatch R with
    | (SolveSep _ _ _) => solveSepStep namer_tac
    | (Transform _) => transformStep namer_tac
    | (IntroduceVars _) => introduceVarStep
    | (IntroduceHyp _ _) => introduceHypStep namer_tac
    | (SolveOne _ _) => solveOneStep namer_tac
    end
  | |- ex _ => eexists
  | |- ?φ ∧ ?R => 
    trySolvePurePreSplit φ R
  | |- ?φ =>
    solve [trySolvePure]
  end.

Ltac solveStepsNT namer_tac :=
  first [solveStepNT namer_tac; solveStepsNT namer_tac | idtac].

Ltac solveStep_with mi post_tac := 
  lazymatch goal with
  | |- environments.envs_entails _ (SolveSep _ _ _) =>
    notypeclasses refine (solve_sep.tac_solve_sep_ext_biabd _ mi _ _ _ _ _ _ _ _ _ _ _);
      [post_tac () | idtac..]
  | |- environments.envs_entails _ (SolveOne _ _) =>
    notypeclasses refine (solve_one.tac_solve_one_ext_abd_input _ mi _ _ _ _ _ _ _); 
      [post_tac () | idtac..]
  end.

Tactic Notation "solveStep" "with" constr(name) :=
  let mi := constr:(Some $ INamed name) in
  let post_tac _ := eapply fic_ext_lookup_Some; first reflexivity in
  solveStep_with mi post_tac.

Tactic Notation "solveStep" "with" "empty_hyp_first" :=
  let mi := constr:(@None ident) in
  let post_tac _ := (eapply fic_ext_lookup_empty_first) in
  solveStep_with mi post_tac.

Tactic Notation "solveStep" "with" "ε₀" := solveStep with empty_hyp_first.

Tactic Notation "solveStep" "with" "empty_hyp_last" :=
  let mi := constr:(@None ident) in
  let post_tac _ := (eapply fic_ext_lookup_empty_last) in
  solveStep_with mi post_tac.

Tactic Notation "solveStep" "with" "ε₁" := solveStep with empty_hyp_last.

Ltac solveStep := solveStepNT default_namer.
Ltac solveSteps := solveStepsNT default_namer.