From iris.proofmode Require Import base reduction coq_tactics tactics notation.
From iris.bi Require Import bi telescopes.


(* This file defines a lot of utilities for dealing with stdpp's telescope types.
   The most tricky parts here are extracting the dependency of a function on its telescoping arguments,
   as well as 'merging' two telescopes.
   These can be found below as ExtractDep and TeleConcatType. *)


Notation "'λᵗ' x .. y , e" :=
  (tele_app (TT := TeleS (λ x, .. (TeleS (λ y, TeleO)) .. )) (λ x, .. (λ y, e) .. ))
  (at level 200, x binder, y binder, right associativity,
   format "'[  ' 'λᵗ'  x  ..  y ']' ,  e") : stdpp_scope.

Notation "'[tele_pair' x ; .. ; z ']'" := 
  (TeleS (fun _ : x => .. (TeleS (fun _ : z => TeleO)) ..))
  (format "[tele_pair  '[hv' x ;  .. ;  z ']' ]").

(* this is less broken than stdpp's default tele_arg stuff
  TODO: Make stdpp merge request *)
Notation "'[tele_arg3'  x ; .. ; z ]" :=
  ((TeleArgCons 
      x
      (.. (TeleArgCons z TargO) .. ))).

(** Helper functions on telescopes *)

(* Getting witnesses out of a term with type telescope *)

Definition targ_proj1 {X} {f : X → tele} (a : TeleS f) : X := 
  let (result, _) := a in result.

Definition targ_proj2 {X} {f : X → tele} (a : TeleS f) : f $ targ_proj1 a :=
  match a as t return (f (targ_proj1 t)) with
  | {| tele_arg_head := h; tele_arg_tail := tail |} => tail
  end.

Notation "(.t1)" := (targ_proj1) (at level 100, no associativity).
Notation "(.t2)" := (targ_proj2) (at level 100, no associativity).

Notation "A .t1" := (targ_proj1 A) (at level 2, left associativity, format "A .t1").
Notation "A .t2" := (targ_proj2 A) (at level 2, left associativity, format "A .t2").

Global Arguments targ_proj1 {_ _} !_ /.
Global Arguments targ_proj2 {_ _} !_ /.

Fixpoint tele_arg_rect (P : ∀ TT, tele_arg TT → Type) 
  (base : P TeleO TargO)
  (rec : (∀ T (b : T → tele) x xs, P (b x) xs → P (TeleS b) (TargS x xs)))
  (TT : tele) : 
  ∀ (xs : tele_arg TT), P TT xs :=
  match TT return (∀ (xs : TT), P TT xs) with
  | TeleO => λ unit_el,
    match unit_el with
    | () => base
    end
  | @TeleS Y b => λ xs,
    match xs with
    | {| tele_arg_head := x; tele_arg_tail := tail |} =>
      rec Y b x tail $ tele_arg_rect P base rec (b x) tail
    end
  end.

(* Adding to the tail of a telescope, when the argument is not dependent on the telescope itself *)
(* this is used in the Ltac which extracts dependency on variables *)

Fixpoint TelePost (TT : tele) (X : Type) : tele :=
  match TT with
  | TeleO => TeleS (λ x : X, TeleO)
  | @TeleS Y binder => TeleS (λ y : Y, TelePost (binder y) X)
  end.

Notation TargS' f a b :=
  ((@TeleArgCons _ (λ x, tele_arg (f x)) a b) : (tele_arg (TeleS f))) (f in scope function_scope, only parsing).

Definition TargPost {TT : tele} (targ : TT) {X : Type} (x : X) : TelePost TT X :=
  (tele_arg_rect (λ TT targ, ∀ (X : Type) (x : X), TelePost TT X)
      (λ X x, TargS' (λ _, TeleO) x TargO)
      (λ Y b y targ rec X tt, TargS y (rec X tt))) TT targ X x.

(* Note that to apply these functions the dependency on targ must be explicit, otherwise the type
  the type of the function cannot even be stated. 
  Right, so now we just apply TargPost' multiple times to merge telescopes, right?
  Nope, have to do it all at once: *)

Fixpoint TeleConcatType (TT : tele) : (TT -t> tele) → tele :=
  match TT return (TT -t> tele) → tele with
  | TeleO => λ TT2, TT2
  | @TeleS Y binder => λ f, @TeleS Y (λ y, TeleConcatType (binder y) (f y))
  end.

(* Alternative definition:
Definition TeleConcatType' (TT : tele) (tfun : TT -t> tele) : tele := tele_fold (@TeleS) id tfun. *)

Definition TeleConcatArg : ∀ {TT : tele} (targ : TT)
   (X : TT -t> tele), tele_app X targ → TeleConcatType TT X :=
  tele_arg_rect (λ TT targ, ∀ (X : TT -t> tele), tele_app X targ → TeleConcatType TT X) 
    (λ X x, x)
    (λ Y b y targ rec X tt, TargS y (rec (X y) tt)).
(* is there a targ_fold..? *)

Definition tele_fun_compose' {TT1 TT2 : tele} {T : Type} :
  (TT2 -t> T) → (TT1 -t> TT2) → (TT1 -t> T) :=
  λ t1 t2, tele_bind (compose (tele_app t1) (tele_app t2)).

Global Arguments tele_fun_compose' {!TT1} {!TT2} {_} _ _ /.

Set Universe Polymorphism.

Definition tele_pointwise (TT : tele) {T} : (relation T) → (relation (TT -t> T)) :=
  λ rel f g, ∀.. tt, rel (tele_app f tt) (tele_app g tt).

Global Arguments tele_pointwise TT {_} rel f g /.

Unset Universe Polymorphism.

Typeclasses Opaque tele_pointwise.

(* Helper classes for handling telescopes *)

Class TeleSummarize {T} (rel : relation T) {TT} (tfun : TT -t> T) (t : T)  :=
  tele_summarize : ∀.. (tt : TT), rel t (tele_app tfun tt).

Class CanonicalSummary {T} (rel : relation T) (summarizer : ∀ {TT : tele}, (TT -t> T) → T) :=
  valid_summary (TT : tele) (tfun : TT -t> T) : ∀.. (tt : TT), rel (summarizer tfun) (tele_app tfun tt).

Class ExtractDep {T : Type} {TT : tele} (P : TT -t> T) (TT1 : tele) (P' : TT1 -t> T) (TT2 : tele) (h : TT1 -t> TT2 -t> TT) :=
  extract_dep : (∀.. tt1 tt2, tele_app P' tt1 = tele_app P (tele_app (tele_app h tt1) tt2)).

Class ExtractDep' {T} {TT} (tfun : TT -t> T) (TT1 : tele) (tfun1  : TT1 -t> T) (TT2 : tele) 
    (h : TT1 -t> TT2 -t> TT) (proj1 : TT -t> TT1) (proj2 : TT -t> TT2) := {
  extract_dep_from_dep' : ExtractDep tfun TT1 tfun1 TT2 h;
  extract_dep_projections : (∀.. (tt : TT), tt = tele_app (tele_app h (tele_app proj1 tt)) (tele_app proj2 tt))
}.

Class TransportTeleProp {PROP : bi} `{TTf : A → tele} (a : A) (S : TTf a -t> PROP) (S' : TeleS TTf -t> PROP) := 
  transported_fun_as_seed : ∀ a' tt, (tele_app (S' a') tt ⊣⊢ 
    ∃ (H : a' = a), tele_app (eq_rect_r (λ a'', TTf a'' -t> PROP) S H) tt).

Section tele_lemmas. (* TODO: not sure how much these tele relations are still used. remove? *)

  Lemma tele_fun_compose_eq' {TT1 TT2 : tele} {T : Type} (f : TT2 -t> T) (g : TT1 -t> TT2) x :
    tele_app (tele_fun_compose' f g) x = tele_app f (tele_app g x).
  Proof. unfold tele_fun_compose'. rewrite tele_app_bind. done. Qed.

  Global Instance tele_pointwise_reflexive {TT T} (rel : relation T) :
    Reflexive rel →
    Reflexive (tele_pointwise TT rel).
  Proof.
    move => Hrel tfun /=.
    rewrite tforall_forall => tt.
    apply Hrel.
  Qed.

  Global Instance tele_pointwise_symmetric {TT T} (rel : relation T) :
    Symmetric rel →
    Symmetric (tele_pointwise TT rel).
  Proof.
    move => Hrel tfun1 tfun2 /= /tforall_forall Htfun.
    apply tforall_forall => tt.
    by apply Hrel, Htfun.
  Qed.

  Global Instance tele_pointwise_transitive {TT T} (rel : relation T) :
    Transitive rel →
    Transitive (tele_pointwise TT rel).
  Proof.
    move => Hrel tfun1 tfun2 tfun3 /= /tforall_forall Htfun12 /tforall_forall Htfun23.
    apply tforall_forall => tt.
    by eapply Hrel; [apply Htfun12 | apply Htfun23].
  Qed.

  Global Instance tele_pointwise_rewrite_relation {TT T} (rel : relation T) :
    RewriteRelation rel →
    RewriteRelation (tele_pointwise TT rel).
  Proof. Qed.

  Global Instance pointwise_empty_tele_subrelation_to {T : Type} (rel : relation T) :
    subrelation rel (tele_pointwise [tele] rel) | 50.
  Proof. done. Qed.

  Global Instance pointwise_empty_tele_subrelation_from {T : Type} (rel : relation T) :
    subrelation (tele_pointwise [tele] rel) rel | 50.
  Proof. done. Qed.

  Global Instance tele_pointwise_subrelation_mono {TT T} (rel1 rel2 : relation T) :
    subrelation rel1 rel2 →
    subrelation (tele_pointwise TT rel1) (tele_pointwise TT rel2).
  Proof.
    rewrite /subrelation => Hrel12 f g /tforall_forall Hfg.
    apply tforall_forall => x.
    apply Hrel12, Hfg.
  Qed.

  Global Instance tele_pointwise_flip_subrelation {TT T} (rel1 rel2 : relation T) :
    subrelation rel1 rel2 →
    subrelation (tele_pointwise TT (flip rel1)) (flip $ tele_pointwise TT rel2).
  Proof.
    rewrite /subrelation => Hrel12 f g /tforall_forall Hfg /=.
    apply tforall_forall => x.
    apply Hrel12, Hfg.
  Qed.

  Global Instance tele_pointwise_flip_subrelation2 {TT T} (rel1 rel2 : relation T) :
    subrelation rel1 rel2 →
    subrelation (flip $ tele_pointwise TT rel1) (tele_pointwise TT (flip rel2)).
  Proof.
    rewrite /subrelation => Hrel12 f g /tforall_forall Hfg /=.
    apply tforall_forall => x.
    apply Hrel12, Hfg.
  Qed.

  Lemma tele_pointwise_app {TT T} (rel : relation T) funL funR : 
    tele_pointwise TT rel funL funR →
    ∀ tt, rel (tele_app funL tt) (tele_app funR tt).
  Proof. by rewrite -tforall_forall. Qed.

  Definition dep_eval {A : Type} {B : A → Type} : ∀ (a : A), (∀ (a : A), B a) → B a
    := λ a f, f a. (* these two are best *)

  Definition dep_eval_tele {TT : tele} {B : TT → Prop} : ∀ (tt : TT), (∀.. (tt : TT), B tt) → B tt.
  Proof. move => tt. by rewrite tforall_forall. Defined.

  Lemma tele_summarize_from_canonical {TT} {T} (tfun : TT -t> T) 
    (rel : relation T) (summarizer : ∀ (TT : tele), (TT -t> T) → T) (summary : T) :
    CanonicalSummary rel summarizer →
    TCEq summary (summarizer TT tfun) →
    TeleSummarize rel tfun summary.
  Proof.
    rewrite /CanonicalSummary /TeleSummarize tforall_forall => + -> tt.
    by move => /(dep_eval TT) /(dep_eval tfun) /(dep_eval_tele tt).
  Qed.

  Global Instance summarize_proper {T : Type} {TT : tele} (rel : relation T) :
    Transitive rel →
    Proper (tele_pointwise _ (rel) ==> (flip rel) ==> impl) (TeleSummarize rel (TT := TT)).
  Proof.
    rewrite /TeleSummarize => Hrel tfun tfun' /tele_pointwise_app Htfun t t' Ht
      /tforall_forall HT.
    rewrite tforall_forall => tt.
    eapply Hrel; first done.
    by eapply Hrel.
  Qed.

End tele_lemmas.

Section summarize_instances.

  Global Instance tele_summarize_from_canonical_extract_dep {TT} {T} (tfun : TT -t> T) 
    (rel : relation T) (summarizer : ∀ (TT : tele), (TT -t> T) → T)
    (TT1 TT2 : tele) (tfun' : TT1 -t> T) (h : TT1 -t> TT2 -t> TT) (proj1 : TT -t> TT1) (proj2 : TT -t> TT2):
    CanonicalSummary rel summarizer →
    ExtractDep' tfun TT1 tfun' TT2 h proj1 proj2 →
    TeleSummarize rel tfun (summarizer TT1 tfun').
  Proof.
    rewrite /CanonicalSummary /TeleSummarize !tforall_forall => +
      [/tforall_forall + /tforall_forall +] tt.
    move => + + /(dep_eval tt) ->.
    move => + /(dep_eval (tele_app proj1 tt)) /tforall_forall /(dep_eval (tele_app proj2 tt)) <-.
    by move => /(dep_eval TT1) /(dep_eval tfun') /tforall_forall.
  Qed.

  Global Instance impl_summary : CanonicalSummary impl (λ TT tfun, ∀.. (tt : TT), tele_app tfun tt).
  Proof.
    rewrite /CanonicalSummary => TT tfun.
    rewrite tforall_forall => tt.
    by rewrite tforall_forall /impl /=.
  Qed.

  Global Instance flip_impl_summary : CanonicalSummary (flip impl) (λ TT tfun, ∃.. (tt : TT), tele_app tfun tt).
  Proof.
    rewrite /CanonicalSummary => TT tfun.
    rewrite tforall_forall => tt.
    rewrite texist_exist /impl /=.
    by eexists.
  Qed.

  Global Instance pointwise_summary {B : Type} (rel : relation B) summarizer {A : Type}:
    CanonicalSummary rel summarizer →
    CanonicalSummary (pointwise_relation A rel) (λ TT tfun a, summarizer TT (tele_map (flip apply a) tfun)).
  Proof.
    rewrite /CanonicalSummary => Hrelsum TT tfun.
    rewrite tforall_forall => tt.
    rewrite /pointwise_relation => a.
    specialize (Hrelsum TT (tele_map (flip apply a) tfun)).
    revert Hrelsum.
    rewrite tforall_forall => Hrelsum.
    specialize (Hrelsum tt); revert Hrelsum.
    by rewrite tele_map_app /apply /=.
  Qed.

(* This instance does not apply nicely when there are prepending [tele] -t> in front of our type. 
  We therefore add a global hint to circumvent this*)
  Lemma tele_pointwise_summary {B: Type} (rel : relation B) summarizer {TT : tele}:
    CanonicalSummary rel summarizer →
    CanonicalSummary (tele_pointwise TT rel) (λ TT' tfun, tele_bind (λ tt, summarizer TT' (tele_map (flip tele_app tt) tfun))).
  Proof.
    rewrite /CanonicalSummary => Hrelsum TT' tfun.
    rewrite tforall_forall => tt' /=.
    rewrite tforall_forall => tt.
    rewrite tele_app_bind.
    specialize (Hrelsum TT' (tele_map (flip tele_app tt) tfun)).
    revert Hrelsum.
    rewrite tforall_forall => Hrelsum.
    specialize (Hrelsum tt'); revert Hrelsum.
    by rewrite tele_map_app.
  Qed.

  Section prop_summarize.
    Context {PROP : bi}.

    Global Instance entails_summary : 
      CanonicalSummary (bi_entails (PROP := PROP)) (λ TT tfun, ∀.. (tt : TT), tele_app tfun tt)%I.
    Proof.
      rewrite /CanonicalSummary => TT tfun.
      rewrite tforall_forall => tt.
      rewrite bi_tforall_forall.
      apply bi.forall_elim.
    Qed.

    Global Instance flip_entails_summary : 
      CanonicalSummary (flip $ bi_entails (PROP := PROP)) (λ TT tfun, ∃.. (tt : TT), tele_app tfun tt)%I.
    Proof.
      rewrite /CanonicalSummary => TT tfun.
      rewrite tforall_forall => tt.
      rewrite bi_texist_exist /=.
      apply bi.exist_intro.
    Qed.
  End prop_summarize.

End summarize_instances.

Section other_instances.
  Global Instance inhabited_empty_tele : Inhabited [tele] := populate [tele_arg].

  Global Instance inhabited_teleS_nondep {A} {TT : tele} :
    Inhabited A →
    Inhabited TT →
    Inhabited (TeleS (λ _ : A, TT)) :=
    λ a tt, populate $ TargS inhabitant inhabitant.
  (* this is better than a match definition since it also reduces for an abstract Inhabited *)

End other_instances.

Section bi_helpers.
  Context {PROP : bi}.

  Lemma bi_texist_mono {TT : tele} (P Q : TT → PROP) :
    (∀ tt, P tt ⊢ Q tt) →
    (∃.. tt, P tt) ⊢ (∃.. tt, Q tt).
  Proof.
    rewrite !bi_texist_exist => HPQ.
    by apply bi.exist_mono.
  Qed.

  Lemma bi_tforall_mono {TT : tele} (P Q : TT → PROP) :
    (∀ tt, P tt ⊢ Q tt) →
    (∀.. tt, P tt) ⊢ (∀.. tt, Q tt).
  Proof.
    rewrite !bi_tforall_forall => HPQ.
    by apply bi.forall_mono.
  Qed.

  Lemma bi_texist_elim {TT : tele} P Q :
    (∀ tt, P tt ⊢ Q) →
    (∃.. tt : TT, P tt) ⊢@{PROP} Q.
  Proof.
    rewrite bi_texist_exist => HPQ.
    by apply bi.exist_elim.
  Qed.

  Lemma bi_tforall_elim {TT : tele} {P} tt :
    (∀.. tt : TT, P tt) ⊢@{PROP} P tt.
  Proof.
    by rewrite bi_tforall_forall (bi.forall_elim tt).
  Qed.

  Lemma dependent_transport `{TTf : A → tele} (a : A) (S : TTf a -t> PROP) :
    TransportTeleProp a S (λ a', tele_bind (λ tt, ∃ (H : a' = a), tele_app (eq_rect_r (λ a'', TTf a'' -t> PROP) S H) tt)%I).
  Proof.
    rewrite /TransportTeleProp => a' tt.
    by rewrite tele_app_bind.
  Qed.

  Lemma independent_transport {A} {TT} (a : A) (S : TT -t> PROP) :
    TransportTeleProp (TTf := λ _ : A, TT) a S (λ a', tele_map (bi_and ⌜a' = a⌝%I) S).
  Proof.
    rewrite /TransportTeleProp => a' tt.
    rewrite !tele_map_app.
    apply (anti_symm _).
    - apply bi.pure_elim_l => ->.
      by rewrite -(bi.exist_intro eq_refl).
    - apply bi.exist_elim => H; subst => /=.
      apply bi.and_intro => //.
      by apply bi.pure_intro.
  Qed.

End bi_helpers.

Ltac concretize_tele_arg tele_type targ :=
  lazymatch tele_type with
  | @TeleS ?X ?tfun => 
    let x := fresh "x" in
    let x := open_constr:((_:X)) in
    let fx := fresh "fx" in
    let fx_type := eval cbn in (tfun x) in
    let fx := open_constr:((_ : tele_arg fx_type)) in
    let new_arg := constr:(TargS' tfun x fx) in
    cut (targ = new_arg); [ move => -> /= | reflexivity ]
  | TeleO => 
    cut (targ = [tele_arg]); [ move => -> /= | reflexivity ]
  end.

Ltac proj_solve := 
  lazymatch goal with
  | |- ?goal = tele_app (@tele_app ?tele_type _ ?tfun ?targ1) ?targ2 => 
    concretize_tele_arg tele_type targ1 
  | |- ?goal = @tele_app ?tele_type _ ?tfun ?targ =>
    concretize_tele_arg tele_type targ 
  | |- _ => reflexivity
  end.

Global Hint Extern 4 (ExtractDep' _ _ _ _ _ _ _) =>
  notypeclasses refine (Build_ExtractDep' _ _ _ _ _ _ _ _ _ _ _); 
  [ cbn; iSolveTC 
  | cbn; intros; by repeat proj_solve
] : typeclass_instances.

Global Hint Extern 4 (TeleSummarize eq ?tfun ?summary) => 
  (* Our only hope to prove this is when tfun is in all cases equal to summary: *)
  rewrite /TeleSummarize; cbn; done : typeclass_instances.


Global Hint Extern 4 (@CanonicalSummary ?T (tele_pointwise ?TT ?rel) ?summary) =>
  let Btype := 
    let rel_type := type of rel in
    match rel_type with
    | relation ?B => B
    | ?B → ?B → Prop => B
    end
  in
  change T with (TT -t> Btype);
  simple eapply @tele_pointwise_summary : typeclass_instances.

Global Hint Extern 4 (TransportTeleProp _ _ _) =>
  refine (independent_transport _ _) || refine (dependent_transport _ _) : typeclass_instances.

Inductive prod2 (A B : Type) : Type := 
  pair2 : A → B → prod2 A B.
Notation "A ⋆ B" := (prod2 A B) (at level 50).
Global Arguments pair2 {_ _} _ _.

Definition fst2 {A B : Type} : prod2 A B → A :=
  λ pr, match pr with | pair2 a b => a end.
Definition snd2 {A B : Type} : prod2 A B → B :=
  λ pr, match pr with | pair2 a b => b end.

Notation "pr .s1" := (fst2 pr) (at level 2, left associativity, format "pr .s1").
Notation "pr .s2" := (snd2 pr) (at level 2, left associativity, format "pr .s2").

Global Arguments fst2 {_ _} !_ /.
Global Arguments snd2 {_ _} !_ /.


Ltac get_curried_application_function input_fun rep_app_fun :=
  let rec get_cur_app_fun the_fun :=
    lazymatch the_fun with 
     | (fun g => rep_app_fun _ (@?body g)) =>
       let body_fun_raw := constr:(fun pr => body pr.s1 pr.s2) in
       let body_fun := eval cbv beta in body_fun_raw in
       get_cur_app_fun body_fun 
     | _ => constr:(the_fun)
    end in
  get_cur_app_fun input_fun.

(* Using the native pair type means that this will break apart functions that were provided as functions on a pair type. *)
Ltac get_uncurried_function_and_tele input_fun := 
  let rec uncur_fun the_fun base_tele :=
    lazymatch (type of the_fun) with
    | ?A ⋆ ?B → ?C =>
      let the_fun2 := eval cbn in (fun (a : A) (b : B) => the_fun (pair2 a b)) in
      let new_tele := constr:(TeleS (fun b : B => base_tele)) in
      uncur_fun the_fun2 new_tele
    | ?A → ?B => constr:(pair2 the_fun (TeleS (fun a : A => base_tele)))
    | ?A => constr:(pair2 the_fun [tele])
  end in 
  uncur_fun input_fun TeleO.

Ltac get_curried_function the_fun :=
  let rec get_curried_fun input_fun :=
    match type of input_fun with
    | ?A → ?B → ?C =>
      let new_input_raw := constr:(fun pr => input_fun pr.s1 pr.s2) in
      let new_input := eval cbv beta in new_input_raw in
      get_curried_fun new_input
    | _ => constr:(input_fun)
    end in
  get_curried_fun the_fun.

Ltac get_repeated_application_function_and_tele input_fun rep_app_fun :=
  let curried_fun := get_curried_application_function input_fun rep_app_fun in
  get_uncurried_function_and_tele curried_fun.

Ltac get_uncurried_function input_fun := 
  let uncur_fun_and_tele := get_uncurried_function_and_tele input_fun in
  let only_fun := eval cbn in (fst2 uncur_fun_and_tele) in
  only_fun.

Ltac get_repeated_application_function input_fun rep_app_fun :=
  let curried_fun := get_curried_application_function input_fun rep_app_fun in
  get_uncurried_function curried_fun.

Ltac get_existential_function input_fun :=
  get_repeated_application_function input_fun @ex.

From iris.base_logic.lib Require Import iprop.

Ltac get_biexist_function_and_tele input_fun :=
  let input_fun_type_raw := type of input_fun in
  let input_fun_type := eval cbn in input_fun_type_raw in
  let the_bi := 
    lazymatch input_fun_type with
    | ?A → (bi_car ?PROP) => constr:((PROP))
    | ?A → iProp ?Σ => constr:((iPropI Σ))
    | ?A → ofe_car (iPropO ?Σ) => constr:((iPropI Σ))
    end in
  get_repeated_application_function_and_tele input_fun (@bi_exist the_bi).

Ltac get_function_on_used_vars input_fun codomain_matcher :=
  (* Specifying codomain is required since otherwise we don't know when to stop when the target T 
    has function type. *)
  (* [tele_arg a] notation for telescopic arguments does not work when the type of a contains evars.
     Likewise, TargPost will choke on lambdas if the type contains evars *)
  let rec get_result_fun the_fun embed_fun :=
    let fun_type := type of the_fun in
    match codomain_matcher fun_type with
    | true => constr:(pair2 the_fun [tele_arg])
    | _ =>
    match (type of the_fun) with 
     | ?A1 ⋆ ?A2 → ?CD => 
        match codomain_matcher CD with
        | true =>
          let fun_uncur := get_uncurried_function the_fun in
          constr:(pair2 fun_uncur embed_fun)
        | _ => fail
        end
     | ?A → ?CD => 
        match codomain_matcher CD with
        | true =>
          lazymatch the_fun with
          | (fun _ => ?b) => 
            let new_embed_raw :=
              lazymatch (type of embed_fun) with
              | tele_arg [tele] => constr:(λ a : A, TargS' (λ _ : A, TeleO) a TargO)
              | ?D → tele_arg ?E => constr:(λ (pr : D ⋆ A), @TargPost E (embed_fun pr.s1) A pr.s2)
              end in
            let raw_type := type of new_embed_raw in
            let new_embed_body := eval cbn in new_embed_raw in (* i guess this simplifies TargS' to tele_arg_cons..? *)
            let new_embed := constr:((new_embed_body : raw_type)) in
            constr:(pair2 b new_embed)
          | _ => 
            let new_embed_raw :=
              lazymatch (type of embed_fun) with
              | tele_arg [tele] => constr:(λ (a : A) (_ : unit), TargS' (λ _ : A, TeleO) a TargO)
              | ?D → tele_arg ?E => constr:(λ (a : A) (d : D), @TargPost E (embed_fun d) A a)
              | ?D → ?E → tele_arg ?F => embed_fun
              end in
            let raw_type := type of new_embed_raw in
            let new_embed_body := eval cbn in new_embed_raw in
            let new_embed := constr:((new_embed_body : raw_type)) in
            constr:(pair2 the_fun new_embed)
          end
        | _ => fail
        end
     | ?A → ?B → ?C =>
       lazymatch the_fun with
       | (fun _ b => @?drop_const b) =>
         let new_embed_raw :=
           lazymatch (type of embed_fun) with
           | tele_arg [tele] => constr:(λ (a : A), TargS' (λ _ : A, TeleO) a TargO)
           | ?D → tele_arg ?E => constr:(λ (pr : D ⋆ A), @TargPost E (embed_fun pr.s1) A pr.s2)
           end in
         let raw_type := type of new_embed_raw in
         let new_embed_body := eval cbn in new_embed_raw in 
         let new_embed := constr:((new_embed_body : raw_type)) in
         get_result_fun drop_const new_embed
       | (fun a _ => @?drop_next a) => 
         let new_embed_raw :=
           lazymatch (type of embed_fun) with
           | tele_arg [tele] => constr:(λ (a : A) (b : B), TargS' (λ _ : A, TeleS (λ _ : B, TeleO)) a (
                                                           TargS' (λ _ : B, TeleO) b TargO))
           | ?D → ?E → tele_arg ?F => constr:(λ (d : D) (pr : E ⋆ B), @TargPost F (embed_fun d pr.s1) B pr.s2)
           | ?D → tele_arg ?E => constr:(λ (a : A) (pr : D ⋆ B), @TargPost (TelePost E A) (@TargPost E (embed_fun pr.s1) A a) B pr.s2)
           end in
         let raw_type := type of new_embed_raw in
         let new_embed_body := eval cbn in new_embed_raw in 
         let new_embed := constr:((new_embed_body : raw_type)) in
         get_result_fun drop_next new_embed
       | _ =>
         let new_embed_raw :=
           lazymatch (type of embed_fun) with
            (* first argument has not yet been registered *)
            (* Somehow (λ (pr : A * B) (_ : unit), [tele_arg pr.2; pr.1]) fails to be constructed *)
           | tele_arg [tele] => constr:(λ (pr : A ⋆ B) (_ : unit), (λ (a : A) (b : B), TargS' (λ _ : A, TeleS (λ _ : B, TeleO)) a (
                                                           TargS'(λ _ : B, TeleO) b TargO)) pr.s1 pr.s2)
            (* first argument has been registered, only register second *)
           | ?D → ?E → tele_arg ?F => constr:(λ (pr : D ⋆ B) (e : E), @TargPost F (embed_fun pr.s1 e) B pr.s2)
            (* first argument has not yet been registered *)
           | ?D → tele_arg ?E => constr:(λ (pr : A ⋆ B) (d : D),
              @TargPost (TelePost E A) (@TargPost E (embed_fun d) A pr.s1) B pr.s2)
           end in
         let raw_type := type of new_embed_raw in
         let new_embed_body := eval cbn in new_embed_raw in 
         let new_embed := constr:((new_embed_body : raw_type)) in
         let new_fun := eval cbv beta in (fun pr => the_fun pr.s1 pr.s2) in
         get_result_fun new_fun new_embed
       end
    end
    end in
  let rec uncurry_second embed_fun base_tele :=
    lazymatch (type of embed_fun) with
    | ?A → unit ⋆ ?C → ?D =>
      let new_embed := eval cbn in (fun a c => embed_fun a (pair2 tt c)) in
      let new_tele := constr:(TeleS (fun c : C => base_tele)) in
      constr:(pair2 (pair2 new_embed new_tele) false)
    | ?A → ?B ⋆ ?C → ?D =>
      let new_embed := eval cbn in (fun a b c => embed_fun a (pair2 b c)) in
      let new_tele := constr:(TeleS (fun c : C => base_tele)) in
      uncurry_second new_embed new_tele
    | ?A → unit → ?C => (* This will also result in a unary function! *)
      let new_embed := eval cbn in (fun a => embed_fun a tt) in
      constr:(pair2 (pair2 new_embed base_tele) false)
    | ?A → ?B → ?C => 
      let new_tele := constr:(TeleS (fun b : B => base_tele)) in
      constr:(pair2 (pair2 embed_fun new_tele) false)
    | ?A → ?D => 
      lazymatch get_uncurried_function_and_tele embed_fun with
      | (pair2 ?final_embed ?final_tele) =>
        constr:(pair2 (pair2 final_embed final_tele) true)
      end
    | _ => constr:(pair2 (pair2 embed_fun base_tele) false)
    end in
  match get_result_fun input_fun [tele_arg] with
  | (pair2 ?the_fun ?embed_fun) => 
    match (uncurry_second embed_fun [tele]) with
    | (pair2 (pair2 ?half_unfolded ?TT2) ?short_circuit) =>
      lazymatch short_circuit with
      | false => (* some arguments were used by the extracted fun *)
        match get_uncurried_function_and_tele half_unfolded with
        | (pair2 ?final_embed ?TT1) =>
          constr:(pair2 (pair2 (pair2 TT1 the_fun) TT2) final_embed)
        end
      | true => (* No arguments were used by the extracted fun *)
        constr:(pair2 (pair2 (pair2 [tele] the_fun) TT2) half_unfolded)
      end
    end
  end.

Global Instance extract_dep_empty_tele {T : Type} (P : [tele] -t> T) : ExtractDep P [tele] P [tele] [tele_arg].
Proof. done. Qed.

Global Hint Extern 4 (@ExtractDep ?T ?TT_in ?input_fun ?TT1 ?output_fun ?TT2 ?embed_fun) =>
  (* It may be the case that TT_in is [tele] -t> TT: if so, simpl this first. *)
  simpl;
  lazymatch goal with
  | |- (@ExtractDep ?T (TeleS _) ?input_fun ?TT1 ?output_fun ?TT2 ?embed_fun) =>
  (* Weird stuff happens when codomain is bi_car (iPropI Σ), since functions to that type will
    sometimes get type checked to iProp Σ - for now we do a manual correction. *)
    lazymatch T with
    | bi_car (iPropI ?Σ) =>
      match get_function_on_used_vars input_fun ltac:(fun T' =>
        let T'' := eval simpl in T' in
        match T'' with
        | iProp Σ => true
        | T => true
        | ofe_car (iPropO Σ) => true
        | _ => false
        end) with
      | (pair2 (pair2 (pair2 ?TT1' ?output_fun') ?TT2') ?embed_fun') =>
        unify TT1' TT1; unify output_fun' output_fun; unify TT2' TT2; unify embed_fun' embed_fun;
        constructor
      end
    | _ => 
      match get_function_on_used_vars input_fun ltac:(fun T' => let T'' := eval simpl in T' in match T'' with T => true | _ => false end) with
      | (pair2 (pair2 (pair2 ?TT1' ?output_fun') ?TT2') ?embed_fun') =>
        unify TT1' TT1; unify output_fun' output_fun; unify TT2' TT2; unify embed_fun' embed_fun;
        constructor
      end
    end
  end : typeclass_instances.

(* When done like above, there is no hope of supporting dependent types.
  This is because we cannot drop the first argument if the second argument still depends on it:
  For example for (λ (n : nat) (H : even n) (b : bool), negb b) we cannot drop the first argument.
  Correct approach would walk back from the last argument.
  To do so we would need to first curry the function (using dependent pairs) to
  (λ (pr : Σ (Σ n, even n), bool) => negb (projT2 pr))
  And then while uncurrying split the arguments. *)

Fixpoint tele_fun_dep (TT : tele) : (TT -t> Type) → Type :=
  match TT return ((TT -t> Type) → Type) with
  | TeleO => (λ cod, cod)
  | TeleS b => (λ cod, ∀ x, tele_fun_dep (b x) (cod x))
  end.

Notation "TT -td_r> A" :=
  (tele_fun_dep TT A) (at level 99, A at level 200, right associativity).

Definition tele_appd {TT : tele} {A : TT -t> Type} (f : TT -td_r> A) : ∀ (tt : tele_arg TT), tele_app A tt :=
  λ tt, (tele_arg_rect (λ (TT : tele) (tt : tele_arg TT), ∀ A, (TT -td_r> A) → tele_app A tt)
     (λ A a, a)
     (λ Y b y targ rec X tt, rec (X y) (tt y))) TT tt A f.
Arguments tele_appd {!_ !_} _ !_ /.

Definition tele_dep_bind_raw {TT : tele} {A : TT -t> Type} : (∀ tt, tele_app A tt) → TT -td_r> A :=
  (fix rec TT : ∀ A, (∀ tt, tele_app A tt) → TT -td_r> A :=
    match TT return ∀ A, (∀ tt, tele_app A tt) → TT -td_r> A with
    | TeleO => λ A f, f [tele_arg]
    | TeleS b => λ A f x, rec (b x) (A x) (λ tt', f (TargS x tt'))
    end) TT A.

Lemma tele_appd_tele_dep_bind_raw_eq {TT1 : tele} {A : TT1 -t> Type} (f : (∀ tt1, tele_app A tt1)) (tt1 : TT1) :
  (tele_appd $ tele_dep_bind_raw f) tt1 = f tt1.
Proof.
  revert tt1.
  induction TT1.
  - by apply tforall_forall => /=.
  - apply tforall_forall => /= x.
    apply tforall_forall => tt1.
    apply H.
Qed.

Fixpoint tele_fun_dep_tele (TT : tele) : (TT -t> tele) → Type → Type :=
  match TT return ((TT -t> tele) → Type → Type) with
  | TeleO => (λ cod A, cod -t> A)
  | TeleS b => (λ cod A, ∀ x, tele_fun_dep_tele (b x) (cod x) A)
  end.

Notation "TT1 -td> TT2 -c> A" :=
  (tele_fun_dep_tele TT1 TT2 A) (at level 99, A at level 200, right associativity).

Definition tele_appd_dep {TT1 : tele} {TT2 : TT1 -t> tele} {A : Type} (f : TT1 -td> TT2 -c> A) : ∀ (tt : tele_arg TT1), tele_app TT2 tt -t> A :=
  λ tt, (tele_arg_rect (λ (TT : tele) (tt : tele_arg TT), ∀ TT2, (TT -td> TT2 -c> A) → tele_app TT2 tt -t> A)
    (λ TT2 f, f)
    (λ Y x y targ rec X tt, rec (X y) (tt y))) TT1 tt TT2 f.
Arguments tele_appd_dep {!_ _ _} _ !_ /.

Definition as_tele_fun_dep_raw {TT1 : tele} {TT2 : TT1 -t> tele} {A : Type} (f : TT1 -td> TT2 -c> A) :
   TT1 -td_r> (tele_map (flip tele_fun A) TT2).
Proof.
  apply tele_dep_bind_raw => tt1.
  rewrite tele_map_app /=.
  by apply tele_appd_dep.
Defined.

Definition as_tele_fun_dep_raw' {TT1 : tele} {TT2 : TT1 -t> tele} {A : Type} (f : TT1 -td> TT2 -c> A) :
   TT1 -td_r> (tele_map (flip tele_fun A) TT2) 
    := (fix rec TT1 : ∀ TT2, (TT1 -td> TT2 -c> A) → TT1 -td_r> (tele_map (flip tele_fun A) TT2) :=
        match TT1 with
        | TeleO => λ TT2 f, f
        | TeleS b => λ TT2 f x, rec (b x) (TT2 x) (f x)
        end) TT1 TT2 f.

Lemma tele_arg_inv_eq {TT : tele} (tt : TT) : 
  match TT as TT return TT → Prop with
  | TeleO => λ tt, tt = TargO
  | TeleS f => λ tt, tt = {| tele_arg_head := tt.t1; tele_arg_tail := tt.t2 |}
   end tt.
Proof. induction TT => //=; by destruct tt. Defined.
(* these are Defined, since sometimes we do eq_rect on equalities produced using these lemmas *)

Lemma tele_arg_S_inv_strong_hd_eq `{b : X → tele} (x : TeleS b) : 
  x = {| tele_arg_head := x.t1; tele_arg_tail := x.t2 |}.
 Proof. exact (tele_arg_inv_eq x). Defined.

Set Universe Polymorphism.
Lemma tele_arg_O_inv_eq (tt : TeleO) :
  tt = TargO.
Proof. apply (tele_arg_inv_eq tt). Defined.

Global Arguments tele_arg_O_inv_eq _ /.
Global Arguments tele_arg_S_inv_strong_hd_eq {_ _} _/.

Lemma unfold_tele_app_S {A C : Type} {TTf : A → tele} (c : TeleS TTf) (g : TeleS TTf -t> C) :
  tele_app (TT := TeleS TTf) g c = tele_app (g c.t1) c.t2.
Proof. destruct c => //=. Defined.

Lemma unfold_tele_app_O {C : Type} (g : TeleO -t> C) (c : TeleO):
  tele_app (TT := TeleO) g c = g.
Proof. rewrite (tele_arg_O_inv_eq c) => //=. Defined.
Unset Universe Polymorphism.

Definition elim_dep_teleo_fun {H : TeleO → Type} : 
  H TargO →
  (∀ (x : TeleO), H x). (* why not use tele_bind? well, dependent functions are a thing *)
Proof. intros. by erewrite tele_arg_O_inv_eq. Defined. (* this makes the construction compute *)

Definition elim_dep_teles_fun `{TTf : A → tele} {H : TeleS TTf → Type} :
  (∀ a, ∀ cons, H $ TargS a cons) →
  (∀ (x : TeleS TTf), H x).
Proof. intros. destruct x => //=. Defined.

Global Arguments elim_dep_teleo_fun {_} _ /.
Global Arguments elim_dep_teles_fun {_ _ _} _ /.

Lemma reducing_tforall_forall {TT : tele} (Ψ : TT → Prop) :
  tforall Ψ → (∀ x, Ψ x).
Proof.
  unfold tforall. induction TT as [|X ft IH].
  - simpl.
    intros ? p. rewrite (tele_arg_O_inv_eq p). done.
  - simpl.
    move => Hx a. destruct a as [x t'].
    specialize (IH x (λ t', Ψ $ TargS x t')).
    eapply IH; last done.
Defined.

Lemma reducing_forall_tforall {TT : tele} (Ψ : TT → Prop) :
  (∀ x, Ψ x) → tforall Ψ.
Proof.
  unfold tforall. induction TT as [|X ft IH] => //=.
  simpl.
  move => Hx a.
  eapply IH => s.
  eapply Hx.
Defined.

Definition tele_app_on_tele_map_tele_fun {TT1 : tele} {TT2 : TT1 -t> tele} {A : Type} {tt1 : TT1} :
  (tele_app (tele_map (flip tele_fun A) TT2) tt1) → ∀ (tt2 : tele_app TT2 tt1), A :=
    (tele_arg_rect (λ TT tt, ∀ TT2, (tele_app (tele_map (flip tele_fun A) TT2) tt) → ∀ (tt2 : tele_app TT2 tt), A)
      (λ X x, λ u, tele_app x u)
      (λ Y b y targ rec TT2 f tt2, rec (TT2 y) f tt2)) TT1 tt1 TT2.

Lemma as_tele_fun_dep_raw_eq {TT1 : tele} {TT2 : TT1 -t> tele} {A : Type} (f : TT1 -td> TT2 -c> A) :
  ∀ tt1, tele_appd_dep f tt1 = eq_rect _ (λ U, U) (tele_appd (as_tele_fun_dep_raw f) tt1) _ (tele_map_app _ _ _).
Proof.
  move => tt1 /=.
  rewrite tele_appd_tele_dep_bind_raw_eq.
  by rewrite rew_opp_r.
Qed.

Lemma as_tele_fun_dep_raw_eq' {TT1 : tele} {TT2 : TT1 -t> tele} {A : Type} (f : TT1 -td> TT2 -c> A) :
  ∀ tt1 tt2, tele_app (tele_appd_dep f tt1) tt2 = tele_app_on_tele_map_tele_fun (tele_appd (as_tele_fun_dep_raw' f) tt1) tt2.
Proof. 
  induction TT1.
  - case => //=.
  - case => y tl tt2. rewrite H //.
Qed.

Definition tele_uncurry {TT1 : tele} {TT2 : TT1 -t> tele} {A : Type} (f : TeleConcatType TT1 TT2 -t> A) 
      : (TT1 -td> TT2 -c> A)
  := (fix rec TT : (∀ TT2, (TeleConcatType TT TT2 -t> A) → TT -td> TT2 -c> A) :=
      match TT return (∀ TT2, (TeleConcatType TT TT2 -t> A) → TT -td> TT2 -c> A) with
      | TeleO => (λ TT2 f, f)
      | TeleS b => (λ TT2 f x, rec (b x) (TT2 x) (f x))
      end) TT1 TT2 f.

Definition tele_curry {TT1 : tele} {TT2 : TT1 -t> tele} {A : Type} (f : TT1 -td> TT2 -c> A)
      : TeleConcatType TT1 TT2 -t> A
  := (fix rec TT : (∀ TT2, (TT -td> TT2 -c> A) → TeleConcatType TT TT2 -t> A) :=
      match TT return (∀ TT2, (TT -td> TT2 -c> A) → TeleConcatType TT TT2 -t> A) with
      | TeleO => (λ TT2 f, f)
      | TeleS b => (λ TT2 f x, rec (b x) (TT2 x) (f x))
      end) TT1 TT2 f.

Lemma tele_curry_uncurry_compose {TT1 : tele} {TT2 : TT1 -t> tele} {A : Type} (f : TeleConcatType TT1 TT2 -t> A) :
  ∀ (tt : TeleConcatType TT1 TT2), tele_app f tt = tele_app (tele_curry $ tele_uncurry f) tt.
Proof.
  induction TT1 => //=.
  apply tforall_forall => /= x.
  apply tforall_forall => x0.
  apply H.
Qed.

Lemma tele_uncurry_curry_compose {TT1 : tele} {TT2 : TT1 -t> tele} {A : Type} (f : TT1 -td> TT2 -c> A) :
  ∀ (tt1 : TT1) (tt2 : tele_app TT2 tt1), 
    tele_app (tele_appd_dep f tt1) tt2 = tele_app (tele_appd_dep (tele_uncurry $ tele_curry f) tt1) tt2.
Proof.
  induction TT1 => //.
  move => tt1 x.
  destruct (tele_arg_S_inv tt1) as [x' [tt1' Htt1]].
  subst. (* <3 subst *)
  apply H.
Qed.

Lemma tele_curry_uncurry_relation {TT1 : tele} {TT2 : TT1 -t> tele} {A : Type} (f : TeleConcatType TT1 TT2 -t> A) :
  ∀ (tt1 : TT1) (tt2 : tele_app TT2 tt1), 
    tele_app f (TeleConcatArg tt1 TT2 tt2) = tele_app (tele_appd_dep (tele_uncurry f) tt1) tt2.
Proof.
  induction TT1 => //.
  { rewrite -tforall_forall //=. }
  rewrite -tforall_forall /= => x.
  apply tforall_forall => tt1 tt2.
  apply H.
Qed.

Definition tele_dep_bind {TT1 : tele} {TT2 : TT1 -t> tele} {A : Type} : (∀ tt1, tele_app TT2 tt1 -t> A) → TT1 -td> TT2 -c> A
  := (fix rec TT : ∀ TT2, (∀ tt1, tele_app TT2 tt1 -t> A) → TT -td> TT2 -c> A :=
      match TT return ∀ TT2, (∀ tt1, tele_app TT2 tt1 -t> A) → TT -td> TT2 -c> A with
      | TeleO => (λ TT2 f, f TargO)
      | TeleS b => (λ TT2 f x, rec (b x) (TT2 x) (λ tt1, f (TargS x tt1)))
      end) TT1 TT2 .

Lemma tele_dep_appd_bind {TT1 : tele} {TT2 : TT1 -t> tele} {A : Type} (f : (∀ tt1, tele_app TT2 tt1 -t> A)) (tt1 : TT1) :
  (tele_appd_dep $ tele_dep_bind f) tt1 = f tt1.
Proof.
  revert tt1.
  induction TT1.
  - by apply tforall_forall => /=.
  - apply tforall_forall => /= x.
    apply tforall_forall => tt1.
    apply H.
Qed.

Record tele_arg_app (TT1 : tele) (TT2 : TT1 -t> tele) : Type := TeleArgApp 
  { tele_arg_start : TT1;
    tele_arg_end : tele_app TT2 tele_arg_start }.
Global Arguments TeleArgApp {_ _} _ _. 

Definition tele_arg_concat_inv_strong_hd {TT1 : tele} {TT2 : TT1 -t> tele} (tt : TeleConcatType TT1 TT2) : tele_arg_app TT1 TT2.
Proof.
  induction TT1.
  - unshelve esplit; [exact TargO| exact tt].
  - destruct tt as [x tt']. specialize (X0 x _ tt') as [tt1 tt2].
    unshelve esplit; first unshelve esplit.
    * exact x. 
    * exact tt1.
    * exact tt2.
Defined.

Lemma tele_arg_concat_inv_eq {TT1 : tele} {TT2 : TT1 -t> tele} (tt : TeleConcatType TT1 TT2) : 
  tt = match (tele_arg_concat_inv_strong_hd tt) with
       | TeleArgApp tt1 tt2 => TeleConcatArg tt1 _ tt2
       end.
Proof.
  induction TT1 => //.
  revert tt => /=.
  apply reducing_tforall_forall => /= x.
  apply reducing_forall_tforall => tt1.
  specialize (H _ _ tt1).
  revert H.
  generalize (tele_arg_concat_inv_strong_hd tt1).
  case => tt1_L tt1_R -> //=.
Defined.

Lemma tele_arg_concat_eq {TT1 : tele} {TT2 : TT1 -t> tele} (tt1 : TT1) (tt2 : tele_app TT2 tt1) :
  TeleArgApp tt1 tt2 = tele_arg_concat_inv_strong_hd (TeleConcatArg tt1 _ tt2).
Proof.
  induction TT1 => //=.
  { set H := tele_arg_O_inv_eq tt1; by subst. }
  revert tt1 tt2 => /=.
  rewrite -tforall_forall => /= x.
  apply reducing_forall_tforall => tt1 tt2.
  by rewrite -(H x (TT2 x) tt1 tt2).
Qed.

Definition TelePairType (TT1 TT2 : tele) := TeleConcatType TT1 (tele_bind (λ _, TT2)).

Definition as_dependent_fun (TT1 TT2 : tele) {A : Type} (f : TT2 -t> A) : 
    ∀ tt1 : TT1, tele_app (tele_bind (λ _, TT2)) tt1 -t> A
  := λ tt, (tele_arg_rect (λ TT tt, ∀ TT2, (TT2 -t> A) → tele_app (tele_bind (λ _, TT2)) tt -t> A)
      (λ TT f, f)
      (λ Y b y targ rec TT2 f, rec TT2 f)) TT1 tt TT2 f.

Definition as_dependent_arg (TT1 TT2 : tele) (tt2 : TT2) : ∀ tt1 : TT1, tele_arg $ tele_app (tele_bind (λ _, TT2)) tt1
:= λ tt1, (tele_arg_rect (λ TT tt1, tele_arg $ tele_app (tele_bind (λ _, TT2)) tt1)
      (tt2)
      (λ Y b y targ rec, rec)) TT1 tt1.

Definition as_non_dependent_arg (TT1 TT2 : tele) : ∀ tt1 : TT1, tele_app (tele_bind (λ _, TT2)) tt1 → TT2
 := λ tt1, (tele_arg_rect (λ TT tt1, tele_app (tele_bind (λ _, TT2)) tt1 → TT2)
      (λ tin, tin)
      (λ Y b y targ rec, rec)) TT1 tt1.

Lemma non_dependent_arg_eq (TT1 TT2 : tele) (tt2 : TT2) : 
    ∀ tt1 : TT1, as_non_dependent_arg _ _ tt1 $ as_dependent_arg _ _ tt2 tt1 = tt2.
Proof. induction TT1; case => //=. Qed.

Lemma dependent_arg_eq (TT1 TT2 : tele) : 
    ∀ (tt1 : TT1) (tt2 : tele_app (tele_bind (λ _, TT2)) tt1), 
      as_dependent_arg _ _ (as_non_dependent_arg _ _ tt1 tt2) tt1 = tt2.
Proof. induction TT1; case => //=. Qed.

Definition tele_pair_arg {TT1 TT2 : tele} (tt1 : TT1) (tt2 : TT2) : TelePairType TT1 TT2
  := TeleConcatArg tt1 (tele_bind (λ _, TT2)) (as_dependent_arg _ _ tt2 tt1).

Lemma dependent_fun_eq (TT1 TT2 : tele) {A : Type} (f : TT2 -t> A) :
  ∀ tt1 tt2, tele_app f tt2 = tele_app ((as_dependent_fun _ _ f) tt1) (as_dependent_arg TT1 TT2 tt2 tt1).
Proof. induction TT1; case => //=. Qed.


(* sometimes we need to use telescopic equalities to get a good type. this helps constructing those fixes *)

Definition tele_fix_app_bind_1_gen {A : tele} {C : Type} (g : C → Type) (a : A) :
  ∀ (TTf : A → C), 
    g (TTf a) →
    g (tele_app (tele_bind TTf) a) :=
  (tele_arg_rect (λ A a, ∀ (TTf : A → C), g (TTf a) → g (tele_app (tele_bind TTf) a))
    (λ TTf w, w)
    (λ Y b y targ rec TTf w, rec (λ b, TTf $ TargS y b) w)) A a.

Global Arguments tele_fix_app_bind_1_gen {_ _} _ !_ _ _ /.

Definition tele_fix_app_bind_2_gen {A : tele} {C : Type} (g : C → Type) (a : A) :
  ∀ (TTf : A → C), 
    g (tele_app (tele_bind TTf) a) →
    g (TTf a) :=
  (tele_arg_rect (λ A a, ∀ (TTf : A → C), g (tele_app (tele_bind TTf) a) → g (TTf a))
    (λ f w, w)
    (λ Y b y targ rec f w, rec (λ b, f $ TargS y b) w)) A a.

Global Arguments tele_fix_app_bind_2_gen {_ _} _ !_ _ _ /.

Lemma tele_fix_app_bind_cancel_1 {A : tele} {C : Type} (g : C → Type) (a : A) (TTf : A → C) w :
  tele_fix_app_bind_2_gen g a TTf $ tele_fix_app_bind_1_gen g a TTf w = w.
Proof.
  induction A; destruct a => //=.
  rewrite /tele_fix_app_bind_2_gen /tele_fix_app_bind_1_gen /= in H.
  rewrite H //. 
Qed.

Lemma tele_fix_app_bind_cancel_2 {A : tele} {C : Type} (g : C → Type) (a : A) (TTf : A → C) w :
  tele_fix_app_bind_1_gen g a TTf $ tele_fix_app_bind_2_gen g a TTf w = w.
Proof.
  induction A; destruct a => //=.
  rewrite /tele_fix_app_bind_2_gen /tele_fix_app_bind_1_gen /= in H.
  rewrite H //.
Qed.

Lemma tele_fix_app_bind_1_outward {A : tele} {C : Type} (G H : C → Type) (a : A) (TTf : A → C) w (h : ∀ c, G c → H c) :
  h _ $ tele_fix_app_bind_1_gen G a TTf w = tele_fix_app_bind_1_gen H a TTf (h _ w).
Proof.
  induction A; destruct a => //=.
  rewrite /tele_fix_app_bind_2_gen /tele_fix_app_bind_1_gen /= in H0.
  rewrite H0 //.
Qed.

Lemma tele_fix_app_bind_2_outward {A : tele} {C : Type} (G H : C → Type) (a : A) (TTf : A → C) w (h : ∀ c, G c → H c) :
  h _ $ tele_fix_app_bind_2_gen G a TTf w = tele_fix_app_bind_2_gen H a TTf (h _ w).
Proof.
  induction A; destruct a => //=.
  rewrite /tele_fix_app_bind_2_gen /tele_fix_app_bind_1_gen /= in H0.
  rewrite H0 //.
Qed.

Lemma tele_fix_app_bind_adjoint {A : tele} {C D : Type} (G : C → Type) (a : A) (TTf : A → C) w (h : G (TTf a) → D) :
  h $ tele_fix_app_bind_2_gen G a TTf w = tele_fix_app_bind_1_gen (λ c, G c → D) a TTf h w.
Proof.
  induction A; destruct a => //=.
  rewrite /tele_fix_app_bind_2_gen /tele_fix_app_bind_1_gen /= in H.
  rewrite H //.
Qed.


Definition tele_app_bind_type_fix {A : tele} {B : Type} (a : A) :
  ∀ (TTf : A → tele), 
  (TTf a -t> B) →
  tele_app (tele_bind TTf) a -t> B := tele_fix_app_bind_1_gen (λ TT : tele, TT -t> B) a.

Global Arguments tele_app_bind_type_fix {_ _ !_ _} _ /.

Definition tele_app_bind_arg_fix {A : tele} (a : A) :
  ∀ (TTf : A → tele),
  tele_app (tele_bind TTf) a → TTf a := tele_fix_app_bind_2_gen tele_arg a.

Global Arguments tele_app_bind_arg_fix {_ !_ _} _ /.

Lemma tele_app_bind_fix_eq {A : tele} {B : Type} {TTf : A → tele} (a : A) arg (f : TTf a -t> B) :
  tele_app (tele_app_bind_type_fix f) arg = tele_app f (tele_app_bind_arg_fix arg).
Proof.
  rewrite tele_fix_app_bind_adjoint /tele_app_bind_type_fix.
  rewrite (tele_fix_app_bind_1_outward (λ TT, TT -t> B) _ _ _ _ (λ c f, tele_app f)) //.
  (* alternatively: induction a => //=. rewrite IHa //. *)
Qed.

Fixpoint tele_eq TT : TT -t> TT -t> Prop :=
  match TT with
  | TeleO => True
  | @TeleS X b => (λ x1 : X, tele_map (λ f x2, 
   tele_bind (λ tt2, ∃ (H : x2 = x1), tele_app f (eq_rect x2 b tt2 _ H))) (tele_eq (b x1)))
  end.

Lemma tele_eq_eq_1 {TT : tele} (y y' : TT) :
  tele_app (tele_app (tele_eq TT) y) y' → y = y'.
Proof.
  induction TT; destruct y.
  - rewrite (tele_arg_O_inv y') //.
  - destruct (tele_arg_S_inv y') as [y1 [y2 ->]] => /=.
    rewrite tele_map_app /=.
    rewrite tele_app_bind /=.
    case => H'. subst.
    move => /H -> //.
Qed.

Lemma tele_eq_eq_2 {TT : tele} (y y' : TT) :
  y = y' → tele_app (tele_app (tele_eq TT) y) y'.
Proof.
  move => -> {y}.
  induction TT; destruct y' => //=.
  rewrite tele_map_app tele_app_bind.
  exists eq_refl => /=.
  by apply H.
Qed.

Lemma tele_eq_app_both {TT : tele} (tt : TT) : tele_app (tele_app (tele_eq TT) tt) tt.
Proof. by apply tele_eq_eq_2. Qed.

Class SimplTeleEq (TT : tele) (teq : TT -t> TT -t> Prop) :=
  teq_spec : (∀.. tt1 tt2, tele_app (tele_app teq tt1) tt2 ↔ tele_app (tele_app (tele_eq TT) tt1) tt2).

Global Instance simpl_teleO_eq : SimplTeleEq [tele] True | 50.
Proof. rewrite /SimplTeleEq //=. Qed.

Global Instance simpl_teleS_nondep_eq (A : Type) (TT : tele) teq : 
  SimplTeleEq TT teq →
  SimplTeleEq (TeleS (λ _ : A, TT)) (λ a1, tele_bind (λ tt1, λ a2, tele_bind (λ tt2, a1 = a2 ∧ tele_app (tele_app teq tt1) tt2))) | 50.
Proof.
  rewrite /SimplTeleEq => /tforall_forall Hteq /= a1.
  apply tforall_forall => tt1 a2.
  apply tforall_forall => tt2.
  rewrite !tele_map_app !tele_app_bind.
  specialize (Hteq tt1). 
  apply (dep_eval_tele tt2) in Hteq; simpl in Hteq.
  split.
  - case => H /Hteq Heq. subst. by exists eq_refl.
  - case => H. subst => /Hteq Heq. eauto.
Qed.

Global Instance simpl_teleS_dep_eq (A : Type) (TTf : A → tele) teq :
  (∀ a, SimplTeleEq (TTf a) (teq a)) →
  SimplTeleEq (TeleS TTf) (λ a1, tele_bind (λ tt1, λ a2, tele_bind (λ tt2, ∃ (H : a2 = a1), tele_app (tele_app (teq a1) tt1) (eq_rect a2 TTf tt2 _ H)))) | 100.
Proof.
  rewrite /SimplTeleEq => + /= a1.
  move => /(dep_eval a1) Heq.
  apply tforall_forall => tt1 a2.
  apply tforall_forall => tt2.
  revert Heq; move => /(dep_eval_tele tt1) Heq.
  rewrite !tele_map_app !tele_app_bind.
  split.
  - case => H. 
    subst => /=. 
    apply (dep_eval_tele tt2) in Heq. simpl in Heq.
    move => /Heq Ht.
    by exists eq_refl.
  - case => H. 
    subst => /=. 
    apply (dep_eval_tele tt2) in Heq. simpl in Heq.
    move => /Heq Ht.
    by exists eq_refl.
Qed.

Global Hint Extern 0 (tforall _) =>
  progress cbn [TeleConcatType tele_appd_dep] : typeclass_instances.


Class TCTForall {TT : tele} (P : TT → Prop) := 
  tctforall : ∀.. (tt : TT), P tt
.

Global Hint Mode TCTForall ! ! : typeclass_instances.

Global Hint Extern 4 (TCTForall _) =>
  unfold TCTForall;
  progress cbn [tforall tele_fold tele_bind tele_app TelePairType TeleConcatType tele_appd_dep];
  intros : typeclass_instances.


Notation "'TC∀..' x .. y , P" := (TCTForall (λ x, .. (TCTForall (λ y, P)) .. ))
  (at level 200, x binder, y binder, right associativity,
  format "TC∀..  x  ..  y ,  P") : stdpp_scope.










