From diaframe.heap_lang Require Import proof_automation.
From diaframe.heap_lang.examples.logatom Require Import atomic_ticketlock_client atomic_ticketlock.
From iris.heap_lang Require Import adequacy notation atomic_heap.

(** Getting a closed proof out of functions parametrized over an atomic heap is somewhat tricky.
  First, we specify that we want to use the default atomic heap implementation: *)
Local Existing Instance primitive_atomic_heap.

(** We want to state a result about a single program expression. If one runs [Check run_foo],
  note that this actually is parametric in the used [atomic_heap]. This is because different
  [atomic_heap]s will give different implementations of load, store, CAS, etc. We have specified
  that we want to use the [primitive_atomic_heap], but the dependency is still there in the type
  of [run_foo]. We use some Ltac tricks to extract the specific [val] we get with [primitive_atomic_heap].
  Below Ltac does not work when one does not make [primitive_atomic_heap] an [Instance], like above *)
Definition run_foo_inst : val :=
  ltac:(
    let anon_func := constr:(λ (Σ : gFunctors) (heapGS0 : heapGS Σ), ltac:(let result := eval cbv in run_foo in exact result)) in
    match anon_func with
    | λ _ _, ?res =>
      exact res
    end
  ).

(** Below shows that we are not doing anything strange with above Ltac. *)
Lemma run_foo_inst_same `{!heapGS Σ} : run_foo_inst = run_foo.
Proof. reflexivity. Qed.

(** We register the specification for [run_foo_inst], since Diaframe will not 
    notice [run_foo] and [run_foo_inst] are unifiable. *)
Global Instance run_foo2_spec `{!heapGS Σ, !tlockG Σ} :
  SPEC {{ True }} run_foo_inst #() {{ RET #(); True }}.
Proof. exact run_foo_spec. Qed.


(** Now we can construct a closed proof: first, we need a specific [gFunctors], which contains
  access to all required (ghost) resources. *)
Definition clientΣ : gFunctors := #[ heapΣ; tlockΣ ].

(** Now use Iris's adequacy theorem. *)
Lemma run_foo_adequate σ : adequate NotStuck (run_foo_inst #()) σ (λ v _, (v = #())).
Proof. apply (heap_adequacy clientΣ)=> ?. iStepsS. Qed.

(** Use: 

Print adequate.

to inspect the meaning of this statement: it contains no reference to anything Diaframe/Iris/separation logic
specific anymore, it just states something about the semantics of executing [run_foo_inst].

Use

Print Assumptions run_foo_adequate.

to check no Axioms were used. This should print
"Closed under the global context." *)



