From diaframe.heap_lang Require Import proof_automation atomic_specs wp_auto_lob.
From iris.heap_lang Require Import proofmode notation atomic_heap.
From iris.algebra Require Import excl auth list frac agree.
From diaframe.lib Require Import own_hints.

(** This file is an edited version of the Elimination Stack verified by Ralf Jung.
    The original version can be found at the iris-examples repository: 
      https://gitlab.mpi-sws.org/iris/examples *)

Definition stackR := (prodR fracR $ agreeR $ listO valO). (* used to be authUR $ optionUR $ exclR $ listO valO *)
Class stack2G Σ := StackG {
  stack_tokG :> inG Σ $ optionUR $ exclR unitO;
  stack_stateG :> inG Σ stackR;
 }.
Definition stack2Σ : gFunctors :=
  #[GFunctor (optionUR $ exclR unitO); GFunctor stackR].

Section stack.
  Context `{!heapGS Σ, stack2G Σ} {aheap: atomic_heap} (N : namespace).
  Notation iProp := (iProp Σ).

  Let offerN := N .@ "offer".
  Let stackN := N .@ "stack".

  Import atomic_heap.notation.

  (** Code. A stack is a pair of two pointers-to-option-pointers, one for the
  head element (if the stack is non-empty) and for the current offer (if it
  exists).  A stack element is a pair of a value an an optional pointer to the
  next element. *)
  Definition new_stack : val :=
    λ: <>,
      let: "head" := ref NONE in
      let: "offer" := ref NONE in
      ("head", "offer").

  Definition push : val :=
    rec: "push" "stack" "val" :=
      let: "head_old" := !(Fst "stack") in
      let: "head_new" := ref ("val", "head_old") in
      if: CAS (Fst "stack") "head_old" (SOME "head_new") then #() else
      (* the CAS failed due to a race, let's try an offer on the side-channel *)
      let: "state" := ref #0 in
      let: "offer" := ("val", "state") in
      (Snd "stack") <- SOME "offer" ;;
      (* wait to see if anyone takes it *)
      (* okay, enough waiting *)
      (Snd "stack") <- NONE ;;
      if: CAS "state" #0 #2 then
        (* We retracted the offer. Just try the entire thing again. *)
        "push" "stack" "val"
      else
        (* Someone took the offer. We are done. *)
        #().

  Definition pop : val :=
    rec: "pop" "stack" :=
      match: !(Fst "stack") with
        NONE => NONE (* stack empty *)
      | SOME "head_old" =>
        let: "head_old_data" := !"head_old" in
        (* See if we can change the master head pointer *)
        if: CAS (Fst "stack") (SOME "head_old") (Snd "head_old_data") then
          (* That worked! We are done. Return the value. *)
          SOME (Fst "head_old_data")
        else
          (* See if there is an offer on the side-channel *)
          match: !(Snd "stack") with
            NONE =>
            (* Nope, no offer. Just try again. *)
            "pop" "stack"
          | SOME "offer" =>
            (* Try to accept the offer. *)
            if: CAS (Snd "offer") #0 #1 then
              (* Success! We are done. Return the offered value. *)
              SOME (Fst "offer")
            else
              (* Someone else was faster. Just try again. *)
              "pop" "stack"
          end
      end.


  Definition stack_content (γs : gname) (l : list val) : iProp :=
    (own γs ((1/2)%Qp, to_agree l)). (* (own γs (◯ Excl' l))%I. *)
  Global Instance stack_content_timeless γs l : Timeless (stack_content γs l) := _.

  Fixpoint list_inv (l : list val) (rep : val) : iProp :=
    match l with
    | nil => ⌜rep = NONEV⌝
    | v::l => ∃ (ℓ : loc) (rep' : val), ⌜rep = SOMEV #ℓ⌝ ∗
                              ℓ ↦□ (v, rep') ∗ list_inv l rep'
    end%I.

  Global Instance list_inv_pers l rep : Persistent (list_inv l rep).
  Proof. revert rep. induction l; iSolveTC. Qed.

  Global Instance list_inv_timeless l rep : Timeless (list_inv l rep).
  Proof. revert rep. induction l; iSolveTC. Qed.

  Definition offer_inv (st_loc : loc) (v : val) (γo γs : gname) (Q : iProp) : iProp :=
    ∃ st : Z, st_loc ↦ #st ∗
      ((⌜st = 0⌝ ∗ own γo (None) ∗ AU << ∃∃ l, stack_content γs l >> @ ⊤∖↑N, ∅
                                      << stack_content γs (v::l), COMM Q >>) ∨ 
       (⌜st = 1⌝ ∗ (Q ∨ own γo (Excl' ()))) ∨ 
       (⌜st = 2⌝ ∗ own γo (Excl' ()))).

  Definition is_offer (γs : gname) (offer_rep : val) : iProp :=
    ⌜offer_rep = NONEV⌝ ∨ ∃ (v : val) (st_loc : loc),
      ⌜offer_rep = SOMEV (v, #st_loc)⌝ ∗
      ∃ γo Q, inv offerN (offer_inv st_loc v γo γs Q).

  Arguments is_offer γs !offer_rep /.

  Local Instance is_offer_persistent γs offer_rep : Persistent (is_offer γs offer_rep).
  Proof. apply _. Qed.

  Definition stack_inv1 (γs : gname) (head : loc) : iProp :=
    (∃ stack_rep l, 
       head ↦ stack_rep ∗ list_inv l stack_rep ∗ own γs ((1/2)%Qp, to_agree l) ∗ emp%I)%I. 
  (* (∗ emp) allows us some leeway to open additional invariants/atomic updates to get own γs *)

  Definition stack_inv2 (γs : gname) (offer : loc) : iProp :=
    (∃ offer_rep, 
       offer ↦ offer_rep ∗ is_offer γs offer_rep)%I.

  Definition is_stack (γs : gname) (s : val) : iProp :=
    (∃ head offer : loc, ⌜s = (#head, #offer)%V⌝ ∗ inv stackN (stack_inv1 γs head)
                                                 ∗ inv stackN (stack_inv2 γs offer))%I.
  Global Instance is_stack_persistent γs s : Persistent (is_stack γs s) := _.

  Lemma is_list_agree xs ys v : list_inv xs v ∗ list_inv ys v ⊢ ⌜xs = ys⌝.
  Proof.
    revert ys v.
    induction xs.
    { move => [| y ys]; iStepsS. }
    move => [| y ys]; do 2 iStepS.
    iAssert (⌜xs = ys⌝)%I as "->"; last done.
    iApply (IHxs ys x1). iStepsS.
  Qed.

  Global Instance is_list_merge xs v ys :
    MergableConsume (list_inv xs v) (λ p Pin Pout,
      TCAnd (TCEq p true) $
      TCAnd (TCEq Pin (list_inv ys v)) $
             TCEq Pout (list_inv xs v ∗ ⌜xs = ys⌝)%I).
  Proof.
    rewrite /MergableConsume /= => p Pin Pout [-> [-> ->]] /=.
    do 2 iStepS. iApply (is_list_agree _ _ (v)). iStepsS.
  Qed.

  Global Instance biabd_islist_none xs :
    HINT ε₀ ✱ [- ; ⌜xs = []⌝] ⊫ [id]; list_inv xs NONEV ✱ [⌜xs = []⌝].
  Proof. iStepsS. Qed.

  Lemma list_inv_unboxed xs v : list_inv xs v ⊢ ⌜val_is_unboxed v⌝%I.
  Proof. induction xs; iStepsS. Qed.

  Global Instance biabd_islist_keep_unboxed p xs v :
    HINT □⟨p⟩ list_inv xs v ✱ [- ; emp] ⊫ [id]; list_inv xs v ✱ [⌜val_is_unboxed v⌝] | 40.
  Proof.
    iStepS. rewrite bi.intuitionistically_if_elim.
    iAssert ⌜val_is_unboxed v⌝%I as "%". { by iApply list_inv_unboxed. }
    iStepsS.
  Qed.

  Obligation Tactic := program_smash_verify.

  Global Program Instance biabd_islist_some (l : loc) xs :
    HINT ε₁ ✱ [x xs' t; l ↦□ (x, t) ∗ list_inv xs' t ∗ ⌜xs = x :: xs'⌝] ⊫ [id]; 
         list_inv xs (SOMEV #l) ✱ [⌜xs = x :: xs'⌝].

  (** Proofs. *)
  Global Program Instance new_stack_spec :
    SPEC {{ True }} new_stack #() {{ (γs : gname) (s : val), RET s; is_stack γs s ∗ stack_content γs [] }}.

  Global Instance frac_agree_upd γs qin qout mqin mqout l1 l2 :
    TCIf (SolveSepSideCondition (l1 = l2)) False TCTrue →
    (* makes sure we find a better instance if they are unifiable *)
    FracSub 1 qin mqin →
    FracSub 1 qout mqout →
    HINT own γs (qin, to_agree l1) ✱ [l3; match mqin with | Some q' => own γs (q', to_agree l3) | None => ⌜l3 = l1⌝ end] 
          ⊫ [bupd]; own γs (qout, to_agree l2) ✱ [match mqout with | Some q' => own γs (q', to_agree l2) | None => emp end ∗ ⌜l3 = l1⌝] | 55.
  Proof.
    move => _ Hq1 Hq2. iStepS.
    enough (own γs (1%Qp, to_agree l1) ==∗ own γs (1%Qp, to_agree l2)).
    destruct mqin; destruct mqout; iStepS; rewrite Hq1;
    try (rewrite assoc -own_op -pair_op frac_op (comm Qp.add) agree_idemp); rewrite Hq2 H0; iStepsS.
    by apply own_update, cmra_update_exclusive.
  Qed.

  Instance no_lob_on_namespace : do_lob.NoLobGen N := I.

  Global Program Instance push_spec γs (s v : val) :
    SPEC ⟨↑ N⟩ l, << is_stack γs s ¦ stack_content γs l >>
      push s v
    << RET #(); stack_content γs (v :: l) >>.

  Global Instance is_list_decompose xs (l : loc) :
    MergableConsume (list_inv xs (SOMEV #l)) (λ p Pin Pout,
      TCAnd (TCEq p false) $
      TCAnd (TCEq Pin empty_hyp_first) $
             TCEq Pout (∃ x xs' v, ⌜xs = x :: xs'⌝ ∗ l ↦□ (x, v) ∗ list_inv xs' v)%I).
  Proof.
    rewrite /MergableConsume /= => p Pin Pout [-> [-> ->]] /=.
    destruct xs; iStepsS.
  Qed.

  Global Instance pop_spec γs (s : val) :
    SPEC ⟨↑ N⟩ l, << is_stack γs s ¦ stack_content γs l >>
      pop s
    << (v : val), RET v; 
        stack_content γs [] ∗ ⌜v = NONEV⌝ ∗ ⌜l = []⌝ ∨ 
        ∃ w l', stack_content γs l' ∗ ⌜l = w :: l'⌝ ∗ ⌜v = SOMEV w⌝ >>.
  Proof.
    do 26 iStepS; last iSmash. (* run until the load of the head *)
    (* if stack is empty, linearize now *)
    destruct x9; iReIntro "H6".
    - iMod "H2" as "Hl". iReIntro "Hl".
      iPoseProof (diaframe_close_right_quant with "H7") as "H7".
      iSmash.
    - (* run until the CAS is run to pop from the head. Succesful case is easy *)
      do 39 iStepS; [ iSmash | | iStepsS]. (* last iStepsS is to solve AU abort side condition *)
      (* load the offer location and check if an offer is available *)
      do 24 iStepS; first iStepsS. (* no offer, then recurse is easy *)
      iIntros "!>". wp_pure credit: "H£". (* get one credit to strip a later from the possibly contained AU *)
      (* try to accept the offer *)
      do 15 iStepS. (* four cases, each with 2 goals for the AU abort side conditions *)
      (* succeeding CAS is interesting (offer was 0). All other cases are easy (offer was 1, done or accepted, or offer was 2, retracted) *)
      2,3,4,5,6,7,8: iSmash.
      (* to get the post condition: peek at the stack, whatever it is. then quickly push and pull the pending AU's *)
      iStepS. iLeft.
      iMod (inv_acc with "H4") as "HN"; first solve_ndisj.
      iMod (lc_fupd_elim_later with "H£ H11") as ">Hl". (* strip a later from the obtained AU *) 
      iReIntro "Hl". iStepS. iReIntro "HN". iStepS.
      (* TODO: fix iReIntro *)
      iPoseProof (diaframe_close_right with "H11") as "H11".
      unseal_diaframe => /=>. iStepS. (* this enforces logically pushing to the stack *)
      (* and now diaframe figures out automatically that the pop should be done, to finish the proof *)
      iSmash.
  Qed.
End stack.


































