From diaframe.heap_lang Require Import proof_automation wp_auto_lob atomic_specs.
From diaframe.lib Require Import own_hints.
From iris.algebra Require Import auth excl numbers.

From diaframe.heap_lang.examples.logatom Require Import queue_lib.

(* This queue is almost, _but not quite_ the Michael Scott queue. The consistent snapshots have been removed to simplify the verification. *)
Definition new_queue : val := 
  λ: <>, let: "n" := AllocN #2 NONEV in AllocN #2 "n".

Definition dequeue : val :=
  rec: "dequeue" "q" :=
    let: "head" := ! "q" in
    let: "next" := ! ("head" +ₗ #1) in
    match: "next" with
      NONE => NONEV
    | SOME "l" =>
      if: CAS "q" "head" "l" then 
        SOME ! "l"
      else 
        "dequeue" "q"
    end.

Definition enqueue : val :=
  λ: "q" "v",
    let: "node" := AllocN #2 NONEV in
    let: "tailptr" := "q" +ₗ #1 in
    "node" <- "v" ;;
    let: "newtail" := "node" in
    (rec: "set_tail" <> :=
      let: "tail" := (! "tailptr") in
      let: "next" := ! ("tail" +ₗ #1) in
      match: "next" with
        NONE => 
          if: CAS ("tail" +ₗ #1) NONE (SOME "newtail") then 
            CAS "tailptr" "tail" "newtail";; #()
          else 
            "set_tail" #()
      | SOME "l" =>
          CAS "tailptr" "tail" "l";;
          "set_tail" #()
      end) #().

Definition queue_value_list_UR : ucmra := authUR $ max_prefix_listUR $ valO.
Definition queue_head_index_UR : ucmra := authUR $ max_natUR.
Class queueAltG Σ := QueueAltG { 
    queueG1 :> inG Σ queue_value_list_UR ;
    queueG2 :> inG Σ queue_head_index_UR;
    queueG3 :> inG Σ queue_base_UR ;
}.
Definition queueAltΣ : gFunctors := 
  #[GFunctor queue_value_list_UR; GFunctor queue_head_index_UR; GFunctor queue_base_UR].

Global Instance subG_queueAltΣ {Σ} : subG queueAltΣ Σ → queueAltG Σ.
Proof. solve_inG. Qed.

Section queue_atomic.
  Context `{!heapGS Σ, !queueAltG Σ}.
  Set Default Proof Using "Type*".

  Definition is_queue_c_all γlv γll γih (l : loc) (ls : list loc) (vs : list val) (i : nat) : iProp Σ :=
    ∃ (l' : loc) (root : loc), 
      l ↦ #l' ∗ is_queue_base root ls vs ∗ own γlv (● to_max_prefix_list vs) ∗
        ⌜(root :: ls) !! i = Some l'⌝ ∗ own γih (● MaxNat i) ∗ own γll (● to_max_prefix_list (root :: ls)) ∗ ⌜length ls = length vs⌝.

  Definition is_queue_c γlv γll γih (l : loc) (cvs : list val) : iProp Σ :=
    ∃ (ls : list loc) (vs : list val) (i : nat), is_queue_c_all γlv γll γih l ls vs i ∗ ⌜cvs = drop i vs⌝.

  Definition queue_tail_inv (l : loc) γll γit : iProp Σ :=
    ∃ (l'' : loc) (ls : list loc) j, (l +ₗ 1) ↦ #l'' ∗ own γll (◯ to_max_prefix_list ls)
      ∗ ⌜ls !! j = Some l''⌝ ∗ own γit (● MaxNat j) ∗ ⌜NoDup ls⌝.

  Let queue_N := nroot.@"msc_queue".

  Definition is_queue_tail (l : loc) γll γit : iProp Σ := inv queue_N (queue_tail_inv l γll γit).

  Global Instance new_queue_spec :
    SPEC {{ True }} 
      new_queue #() 
    {{ (l : loc) γll γlv γih γit, RET #l; 
      is_queue_c γlv γll γih l [] ∗ is_queue_tail l γll γit }}.
  Proof. (* root is logically separate from the current head, so we need to give an explicit witness *)
    iStepsS. iExists [],[], x. iStepsS.
  Qed.

  Hint Extern 4 (_ ≤ _)%nat => done : solve_pure_add.

  Global Instance dequeue_spec γll γlv γih γit (l : loc) :
    SPEC vs, << is_queue_tail l γll γit ¦ is_queue_c γlv γll γih l vs >>
      dequeue #l 
    << (ov : val), RET ov; ⌜ov = NONEV⌝ ∗ ⌜vs = []⌝ ∗ is_queue_c γlv γll γih l []
                 ∨ ∃ v vs', ⌜vs = v :: vs'⌝ ∗ ⌜ov = SOMEV v⌝ ∗ is_queue_c γlv γll γih l vs' >>.
  Proof.
    iStepsS. (* peek at head *) iLeft. iStepsS.
    - (* queue was empty *) iSmash. all: (rewrite drop_ge //; lia).
    - (* queue was non-empty, abort, try to CAS*) iLeft. iStepsS; last iSmash. (* failing CAS is easy *)
      iRight. (* commit, we need to establish some pure facts*) iStepS. iRight.
      assert (x22 = x11); last subst. (* not strictly needed to proceed, but this pure fact is needed for multiple subgoals *)
      { eapply NoDup_lookup. exact H15. prove_lookup. prove_lookup. }
      erewrite drop_S; last prove_lookup.
      iStepsS.
  Qed.

  Global Instance enqueue_spec γll γlv γih γit (l : loc) (v : val) :
    SPEC vs, << is_queue_tail l γll γit ¦ is_queue_c γlv γll γih l vs >>
      enqueue #l v
    << RET #(); is_queue_c γlv γll γih l (vs ++ [v]) >>.
  Proof.
    iStepsS. (* allocate node, start recursion, peek at tail and at next of tail *)
    - (* next was empty: try to add our element here *)iLeft. iStepS. iLeft. iStepS.
      iStepsS; last iSmash. (* failing CAS is easy *)
      iRight. iStepsS. (* CAS succeeded: commit. We need manual help for the second CAS, to establish some pure facts *)
      { rewrite !app_length /=; lia. }
      { rewrite drop_app_le //. apply lookup_lt_Some in H10. simpl in H10. lia. }
      split. 
      { assert ((x22 :: x17 ++ [x2]) !! S (length x17) = Some x2) as Hd; last by prove_lookup.
        simpl. by apply list_lookup_middle. }
      iStepsS.
      enough (length x1 = x23); first lia.
      eapply (NoDup_lookup (longest x19 (x22 :: x17 ++ [x2]))).
      prove_nodup. prove_lookup. prove_lookup.
    - (* next was SOME: abort and swing tail pointer forward *) iLeft. iStepsS.
      split.
      { assert ((x19 :: x16) !! S x13 = Some x14) as Hd; prove_lookup. }
      iStepsS.
      enough (x13 = x23); first lia.
      eapply (NoDup_lookup (longest x22 (x19 :: x16))).
      prove_nodup. prove_lookup. prove_lookup.
  Qed.
End queue_atomic.

















