From diaframe.heap_lang Require Import proof_automation atomic_specs wp_auto_lob.
From diaframe.lib Require Import own_hints max_prefix_list_hints.
From iris.heap_lang Require Import proofmode.
From iris.algebra Require Import auth excl numbers.

From diaframe.heap_lang.examples.logatom Require Import queue_lib.

Definition new_queue : val := λ: "_", 
  let: "n" := AllocN #2 NONEV in ref "n".

Definition dequeue : val :=
  rec: "dequeue" "q" :=
    let: "head" := ! "q" in
    let: "next" := ! ("head" +ₗ #1) in
    match: "next" with
      NONE => NONEV
    | SOME "l" =>
      if: CAS "q" "head" "l" then 
        SOME ! "l"
      else 
        "dequeue" "q"
    end.

Definition get_tail : val :=
  rec: "get_tail" "head" :=
    let: "next" := ! ("head" +ₗ #1) in
    match: "next" with
      NONE => "head"
    | SOME "l" =>
      "get_tail" "l"
    end.

Definition set_tail : val :=
  rec: "set_tail" "head" "next" :=
    let: "tail" := get_tail "head" in
    if: CAS ("tail" +ₗ #1) NONE (SOME "next") then 
      #()
    else 
      "set_tail" "head" "next". 
  (* doing the recursive call with "head" is inefficient, but most similar to Capers implementation. 
     One can freely replace "head" with "tail" to get a more efficient version that can be verified without changes to the proof *)

Definition enqueue : val :=
  λ: "q" "v",
    let: "node" := AllocN #2 NONEV in
    "node" <- "v" ;;
    let: "head" := ! "q" in
    set_tail "head" "node".

Definition queue_value_list_UR : ucmra := authUR $ max_prefix_listUR $ valO.
Definition queue_head_index_UR : ucmra := authUR $ max_natUR.
Class queueAltG Σ := QueueAltG { 
    queueG1 :> inG Σ queue_value_list_UR ;
    queueG2 :> inG Σ queue_head_index_UR;
    queueG3 :> inG Σ queue_base_UR ;
}.
Definition queueAltΣ : gFunctors := 
  #[GFunctor queue_value_list_UR; GFunctor queue_head_index_UR; GFunctor queue_base_UR].

Global Instance subG_queueAltΣ {Σ} : subG queueAltΣ Σ → queueAltG Σ.
Proof. solve_inG. Qed.

Section queue_atomic.
  Context `{!heapGS Σ, !queueAltG Σ}.

  Definition is_queue_c_all γlv γll γi (l : loc) (ls : list loc) (vs : list val) (i : nat) : iProp Σ :=
    ∃ (l' : loc) (root : loc), 
      l ↦ #l' ∗ is_queue_base root ls vs ∗ own γlv (● to_max_prefix_list vs) ∗
        ⌜(root :: ls) !! i = Some l'⌝ ∗ own γi (● MaxNat i) ∗ own γll (● to_max_prefix_list (root :: ls)) ∗ ⌜length ls = length vs⌝.

  Definition is_queue_c γlv γll γi (l : loc) (cvs : list val) : iProp Σ :=
    ∃ (ls : list loc) (vs : list val) (i : nat), is_queue_c_all γlv γll γi l ls vs i ∗ ⌜cvs = drop i vs⌝.

  Lemma new_queue_spec :
    {{{ True }}} new_queue #() {{{ (l : loc) γll γlv γm, RET #l; is_queue_c γll γlv γm l [] }}}.
  Proof. (* root is logically separate from the current head, so we need to give an explicit witness *)
    iStepsS. iExists [],[], x0. iStepsS.
  Qed.

  Hint Extern 4 (_ ≤ _)%nat => done : solve_pure_add.

  Global Instance dequeue_spec γll γlv γm (l : loc) :
    SPEC vs, << is_queue_c γll γlv γm l vs >>
      dequeue #l 
    << (ov : val), RET ov; ⌜ov = NONEV⌝ ∗ ⌜vs = []⌝ ∗ is_queue_c γll γlv γm l []
                 ∨ ∃ v vs', ⌜vs = v :: vs'⌝ ∗ ⌜ov = SOMEV v⌝ ∗ is_queue_c γll γlv γm l vs' >>.
  Proof.
    iStepsS. (* peek at head *) iLeft. iStepsS.
    - (* queue was empty: commit and done*) iSmash. all: (rewrite drop_ge //; lia).
    - (* queue was not empty: abort, try to CAS *)iLeft. iStepsS; last iSmash. (* failing CAS is easy *)
      iRight. (* commit, we need to establish some pure facts *) iStepS. iRight.
      assert (x21 = x10); last subst. (* not strictly needed to proceed, but this pure fact is needed for multiple subgoals *)
      { eapply NoDup_lookup. exact H15. prove_lookup. prove_lookup. }
      erewrite drop_S; last prove_lookup. 
      iStepsS.
  Qed.

  Global Instance get_tail_spec γll γlv γm (l lq : loc) root ls' j :
    SPEC ls vs i, << own γll (◯ (to_max_prefix_list (root :: ls'))) ∗ ⌜(root :: ls') !! j = Some lq⌝ 
                   ¦ is_queue_c_all γlv γll γm l ls vs i >>
      get_tail #lq 
    << (lt : loc), RET #lt; ⌜lt = default root (last ls)⌝ ∗ is_queue_c_all γlv γll γm l ls vs i >>.
  Proof. iStepS. iLöb as "IH" forall (j lq ls'). iSmash. Qed.
    (* TODO: auto-löb doesnt generalize on j! *)

  Global Instance set_tail_spec γll γlv γm (l lq : loc) root (lt' : loc) ls' j :
    SPEC ls vs i v, << own γll (◯ (to_max_prefix_list (root :: ls'))) ∗ ⌜(root :: ls') !! j = Some lq⌝ 
                     ¦ is_queue_c_all γlv γll γm l ls vs i ∗ lt' ↦ v ∗ (lt' +ₗ 1) ↦ NONEV >>
      set_tail #lq #lt'
    << RET #(); is_queue_c_all γlv γll γm l (ls ++ [lt']) (vs ++ [v]) i >>.
  Proof.
    iStepsS; last iSmash. (* iSmash deals with AU abort sidecondition *) iLeft. (* abort, try to CAS *) iStepsS.
    (* Hints are not smart enough to see that the following is always true, so (default x16 (last x1) ∈ (x16 :: x1)) *)
    assert ((x16 :: x1) !! length x1 = Some (default x16 (last x1))).
    { rewrite last_lookup. destruct x1 => //=.
      assert (is_Some ((l0 :: x1) !! length x1)) as [w ->]; last done.
      apply lookup_lt_is_Some_2. simpl; lia. }
    iStepsS; last iSmash. (* failing CAS is easy *)
    iSmash. rewrite !app_length //=. lia.
  Qed.

  Global Instance enqueue_spec γll γlv γm (l : loc) (v : val) :
    SPEC vs, << is_queue_c γll γlv γm l vs >>
      enqueue #l v
    << RET #(); is_queue_c γll γlv γm l (vs ++ [v]) >>.
  Proof.
    iSmash. (* TODO: here we úse the 'feature' that iSmash disregards unsolved pure goals *)
    apply eq_sym, drop_app_le.
    apply lookup_lt_Some in H4. simpl in H4. lia.
  Qed.
End queue_atomic.

















